"""Unit tests for the dashboard.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Set, TYPE_CHECKING

import kgb
from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.db.models import Count, Q
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import (Profile,
                                         ReviewRequestVisit)
from reviewboard.accounts.testing.queries import \
    get_user_local_site_profile_equeries
from reviewboard.datagrids.grids import DashboardDataGrid
from reviewboard.datagrids.tests.base import BaseViewTestCase
from reviewboard.diffviewer.models import DiffSet
from reviewboard.reviews.models import (Group,
                                        ReviewRequest,
                                        ReviewRequestDraft)
from reviewboard.reviews.testing.queries.review_requests import (
    get_review_requests_from_user_q,
    get_review_requests_to_group_q,
    get_review_requests_to_or_from_user_q,
    get_review_requests_to_user_directly_q,
    get_review_requests_to_user_q,
)
from reviewboard.site.models import LocalSite
from reviewboard.site.testing.queries import \
    get_local_site_is_accessible_by_equeries
from reviewboard.testing.queries.http import get_http_request_start_equeries

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries

    from reviewboard.testing.queries.base import ExpectedQResult


class DashboardViewTests(kgb.SpyAgency, BaseViewTestCase):
    """Unit tests for the dashboard view."""

    datagrid_url = '/dashboard/'

    @add_fixtures(['test_users'])
    def test_overview(self):
        """Testing dashboard view (overview)"""
        self._test_overview()

    @add_fixtures(['test_site', 'test_users'])
    def test_overview_with_local_site(self):
        """Testing dashboard view (overview) on a Local Site"""
        self._test_overview(with_local_site=True,
                            local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_overview_with_local_sites_in_db(self):
        """Testing dashboard view (overview) with Local Sites in the database
        """
        self._test_overview(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_incoming(self):
        """Testing dashboard view (Incoming -> Open)"""
        self._test_incoming_open()

    @add_fixtures(['test_site', 'test_users'])
    def test_incoming_with_local_site(self):
        """Testing dashboard view (Incoming -> Open) on a Local Site"""
        self._test_incoming_open(with_local_site=True,
                                 local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_incoming_with_local_sites_in_db(self):
        """Testing dashboard view (Incoming -> Open) with Local Sites in the
        database
        """
        self._test_incoming_open(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_outgoing(self):
        """Testing dashboard view (Outgoing -> Open)"""
        self._test_outgoing_open()

    @add_fixtures(['test_site', 'test_users'])
    def test_outgoing_open_with_local_site(self):
        """Testing dashboard view (Outgoing -> Open) on a Local Site"""
        self._test_outgoing_open(with_local_site=True,
                                 local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_outgoing_open_with_local_sites_in_db(self):
        """Testing dashboard view (Outgoing -> Open) with Local Sites in the
        database
        """
        self._test_outgoing_open(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_outgoing_all(self):
        """Testing dashboard view (Outgoing -> All)"""
        self._test_outgoing_all()

    @add_fixtures(['test_site', 'test_users'])
    def test_outgoing_all_with_local_site(self):
        """Testing dashboard view (Outgoing -> All) on Local Site"""
        self._test_outgoing_all(with_local_site=True,
                                local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_outgoing_all_with_local_sites_in_db(self):
        """Testing dashboard view (Outgoing -> All) with Local Sites in the
        database
        """
        self._test_outgoing_all(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_incoming_to_me(self):
        """Testing dashboard view (Incoming -> To Me)"""
        self._test_to_me()

    @add_fixtures(['test_site', 'test_users'])
    def test_incoming_to_me_with_local_site(self):
        """Testing dashboard view (Incoming -> To Me) on Local Site"""
        self._test_to_me(with_local_site=True,
                         local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_incoming_to_me_with_local_sites_in_db(self):
        """Testing dashboard view (Incoming -> To Me) with Local Sites in the
        database
        """
        self._test_to_me(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_to_group_with_joined_groups(self):
        """Testing dashboard view with to-group and joined groups"""
        self._test_to_group_with_joined_groups()

    @add_fixtures(['test_site', 'test_users'])
    def test_to_group_with_joined_groups_with_local_site(self):
        """Testing dashboard view with to-group and joined groups on
        Local Site
        """
        self._test_to_group_with_joined_groups(with_local_site=True,
                                               local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_to_group_with_joined_groups_with_local_sites_in_db(self):
        """Testing dashboard view with to-group and joined groups on
        Local Sites in the database
        """
        self._test_to_group_with_joined_groups(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_to_group_with_unjoined_public_group(self):
        """Testing dashboard view with to-group and unjoined public group"""
        self._test_to_group_with_unjoined_groups_public()

    @add_fixtures(['test_site', 'test_users'])
    def test_to_group_with_unjoined_public_group_with_local_site(self):
        """Testing dashboard view with to-group and unjoined public group on
        Local Site
        """
        self._test_to_group_with_unjoined_groups_public(
            with_local_site=True,
            local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_to_group_with_unjoined_public_group_with_local_sites_in_db(self):
        """Testing dashboard view with to-group and unjoined public group with
        Local Site in the database
        """
        self._test_to_group_with_unjoined_groups_public(
            local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_to_group_with_unjoined_private_group(self):
        """Testing dashboard view with to-group and unjoined private group"""
        self._test_to_group_with_unjoined_private_group()

    @add_fixtures(['test_site', 'test_users'])
    def test_to_group_with_unjoined_private_group_with_local_site(self):
        """Testing dashboard view with to-group and unjoined private group
        on Local Site
        """
        self._test_to_group_with_unjoined_private_group(
            with_local_site=True,
            local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_to_group_with_unjoined_private_group_with_local_sites_in_db(self):
        """Testing dashboard view with to-group and unjoined private group
        with Local Sites in the database
        """
        self._test_to_group_with_unjoined_private_group(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_with_all_columns(self):
        """Testing dashboard view with all columns"""
        self._test_with_all_columns()

    @add_fixtures(['test_site', 'test_users'])
    def test_with_all_columns_with_local_site(self):
        """Testing dashboard view with all columns on a Local Site"""
        self._test_with_all_columns(with_local_site=True,
                                    local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_with_all_columns_with_local_sites_in_db(self):
        """Testing dashboard view with all columns with Local Sites in the
        database
        """
        self._test_with_all_columns(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_show_archived(self):
        """Testing dashboard view with show-archived"""
        self._test_show_archived()

    @add_fixtures(['test_site', 'test_users'])
    def test_show_archived_with_local_site(self):
        """Testing dashboard view with show-archived on a Local Site"""
        self._test_show_archived(with_local_site=True,
                                 local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_show_archived_with_local_sites_in_db(self):
        """Testing dashboard view with show-archived with Local Sites in the
        database
        """
        self._test_show_archived(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_archived_with_null_extra_data(self):
        """Testing dashboard view with archived review requests and null
        extra_data
        """
        self._test_archived_with_null_extra_data()

    @add_fixtures(['test_site', 'test_users'])
    def test_archived_with_null_extra_data_with_local_site(self):
        """Testing dashboard view with archived review requests and null
        extra_data on a Local Site
        """
        self._test_archived_with_null_extra_data(
            with_local_site=True,
            local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_archived_with_null_extra_data_with_local_sites_in_db(self):
        """Testing dashboard view with archived review requests and null
        extra_data with Local Sites in a database
        """
        self._test_archived_with_null_extra_data(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_sidebar(self):
        """Testing dashboard sidebar"""
        self._test_sidebar()

    @add_fixtures(['test_site', 'test_users'])
    def test_sidebar_with_local_site(self):
        """Testing dashboard sidebar on a Local Site"""
        self._test_sidebar(with_local_site=True,
                           local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_sidebar_with_local_sites_in_db(self):
        """Testing dashboard sidebar with Local Sites in the database"""
        self._test_sidebar(local_sites_in_db=True)

    def _test_overview(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for Overview.

        This accesses the Dashboard, checking the Overview view for queries
        and expected datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        profile.starred_groups.clear()

        grumpy = User.objects.get(username='grumpy')

        review_request_pks: List[int] = []

        for i in range(10):
            if i < 5:
                submitter = user
            else:
                submitter = grumpy

            review_request = self.create_review_request(
                local_site=local_site,
                local_id=i + 1,
                summary='Test %s' % (i + 1),
                submitter=submitter,
                publish=True)

            if i < 5:
                review_request.target_people.add(user)
                review_request_pks.append(review_request.pk)

        # This will be ordered with newest at the top.
        review_request_pks.reverse()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=review_request_pks,
            rows_q_result=get_review_requests_to_or_from_user_q(
                user=user,
                to_or_from_user=user,
                to_or_from_user_profile=profile,
                local_site=local_site,
                has_local_sites_in_db=local_sites_in_db,
                status='P'))

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'view': 'overview',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 5)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 5')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 4')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[3]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[4]['object'].summary, 'Test 1')

    def _test_incoming_open(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for the Incoming -> Open view.

        This accesses the Dashboard, checking the Incoming -> Open view for
        queries and expected datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        profile.starred_groups.clear()

        review_request_pks: List[int] = []

        for i in range(10):
            review_request = self.create_review_request(
                local_site=local_site,
                local_id=i + 1,
                summary='Test %s' % (i + 1),
                publish=True)

            if i < 5:
                review_request.target_people.add(user)
                review_request_pks.append(review_request.pk)

        # This will be ordered with newest at the top.
        review_request_pks.reverse()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=review_request_pks,
            rows_q_result=get_review_requests_to_user_q(
                user=user,
                to_user=user,
                to_user_profile=profile,
                local_site=local_site,
                has_local_sites_in_db=local_sites_in_db,
                status='P'))

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'view': 'incoming',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 5)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 5')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 4')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[3]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[4]['object'].summary, 'Test 1')

    def _test_outgoing_open(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for Outgoing -> Open.

        This accesses the Dashboard, checking the Outgoing -> Open view for
        queries and expected datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='admin', password='admin')

        user = User.objects.get(username='admin')
        grumpy = User.objects.get(username='grumpy')

        profile = user.get_profile()
        profile.starred_groups.clear()

        review_request_pks: List[int] = []

        for i in range(10):
            if i < 5:
                submitter = user
            else:
                submitter = grumpy

            review_request = self.create_review_request(
                local_site=local_site,
                local_id=i + 1,
                summary='Test %s' % (i + 1),
                submitter=submitter,
                publish=True)

            if submitter is user:
                review_request_pks.append(review_request.pk)

        # This will be ordered with newest at the top.
        review_request_pks.reverse()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=review_request_pks,
            extra=self._build_extra(user=user,
                                    include_new_review_count=True),
            include_star_column=True,
            rows_q_result=get_review_requests_from_user_q(
                user=user,
                from_user=user,
                local_site=local_site,
                has_local_sites_in_db=local_sites_in_db,
                status='P'))

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'view': 'outgoing',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 5)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 5')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 4')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[3]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[4]['object'].summary, 'Test 1')

    def _test_outgoing_all(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for Outgoing -> All.

        This accesses the Dashboard, checking the Outgoing -> All view for
        queries and expected datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        profile.starred_groups.clear()

        grumpy = User.objects.get(username='grumpy')

        review_request_pks: List[int] = []

        for i in range(10):
            if i < 5:
                submitter = user
            else:
                submitter = grumpy

            review_request = self.create_review_request(
                summary='Test %s' % (i + 1),
                submitter=submitter,
                local_site=local_site,
                local_id=i + 1,
                publish=True)

            if submitter is user:
                review_request_pks.append(review_request.pk)

        # This will be ordered with newest at the top.
        review_request_pks.reverse()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=review_request_pks,
            rows_q_result=get_review_requests_from_user_q(
                user=user,
                from_user=user,
                local_site=local_site,
                has_local_sites_in_db=local_sites_in_db,
                status=None))

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'view': 'mine',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 5)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 5')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 4')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[3]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[4]['object'].summary, 'Test 1')

    def _test_to_me(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for Incoming -> To Me.

        This accesses the Dashboard, checking the Incoming -> To Me view for
        queries and expected datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        profile.starred_groups.clear()

        group = self.create_review_group(local_site=local_site)
        group.users.add(user)

        review_request_pks: List[int] = []

        for i in range(15):
            review_request = self.create_review_request(
                summary='Test %s' % (i + 1),
                publish=True,
                local_site=local_site,
                local_id=i + 1)

            if i < 5:
                review_request.target_people.add(user)
                review_request_pks.append(review_request.pk)
            elif i < 10:
                review_request.target_groups.add(group)

        # This will be ordered with newest at the top.
        review_request_pks.reverse()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=review_request_pks,
            rows_q_result=get_review_requests_to_user_directly_q(
                user=user,
                to_user=user,
                to_user_profile=profile,
                local_site=local_site,
                status='P',
                has_local_sites_in_db=local_sites_in_db))

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'view': 'to-me',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 5)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 5')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 4')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[3]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[4]['object'].summary, 'Test 1')

    def _test_to_group_with_joined_groups(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for Incoming -> To Group with joined groups.

        This accesses the Dashboard, checking the Incoming -> To Group view
        with joined groups for queries and expected datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        if local_sites_in_db:
            local_site_q = Q(local_site=local_site)
        else:
            local_site_q = Q()

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        profile.starred_groups.clear()

        group1 = self.create_review_group(
            name='devgroup',
            local_site=local_site)
        group1.users.add(User.objects.get(username='doc'))

        group2 = self.create_review_group(
            name='test-group',
            local_site=local_site)

        review_request_pks: List[int] = []

        for i in range(15):
            review_request = self.create_review_request(
                summary='Test %s' % (i + 1),
                publish=True,
                local_site=local_site,
                local_id=i + 1)

            if i < 5:
                review_request.target_groups.add(group1)
                review_request_pks.append(review_request.pk)
            elif i < 10:
                review_request.target_groups.add(group2)

        # This will be ordered with newest at the top.
        review_request_pks.reverse()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        load_state_equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch the group being viewed',
                'model': Group,
                'where': (Q(name='devgroup') &
                          local_site_q),
            },
        ]

        if local_site:
            load_state_equeries += [
                {
                    '__note__': 'Fetch the accessed Local Site (again)',
                    'model': LocalSite,
                    'where': Q(id=local_site.pk),
                },
            ] + get_local_site_is_accessible_by_equeries(
                user=user,
                local_site=local_site)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=review_request_pks,
            load_state_equeries=load_state_equeries,
            rows_q_result=get_review_requests_to_group_q(
                to_group_name='devgroup',
                user=user,
                local_site=local_site,
                status='P',
                has_local_sites_in_db=local_sites_in_db))

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'group': 'devgroup',
                    'view': 'to-group',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 5)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 5')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 4')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[3]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[4]['object'].summary, 'Test 1')

    def _test_to_group_with_unjoined_groups_public(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for Incoming -> To Group with unjoined public groups.

        This accesses the Dashboard, checking the Incoming -> To Group view
        with unjoined public groups for queries and expected datagrid row
        results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        if local_sites_in_db:
            local_site_q = Q(local_site=local_site)
        else:
            local_site_q = Q()

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        profile.starred_groups.clear()

        group = self.create_review_group(
            name='devgroup',
            local_site=local_site)

        review_request = self.create_review_request(
            summary='Test 1',
            publish=True,
            local_site=local_site,
            local_id=1)
        review_request.target_groups.add(group)

        self._prefetch_cached(local_site=local_site,
                              user=user)

        load_state_equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch the group being viewed',
                'model': Group,
                'where': (Q(name='devgroup') &
                          local_site_q),
            },
        ]

        if local_site:
            load_state_equeries += [
                {
                    '__note__': 'Fetch the accessed Local Site (again)',
                    'model': LocalSite,
                    'where': Q(id=local_site.pk),
                },
            ] + get_local_site_is_accessible_by_equeries(
                user=user,
                local_site=local_site)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=[review_request.pk],
            load_state_equeries=load_state_equeries,
            rows_q_result=get_review_requests_to_group_q(
                to_group_name='devgroup',
                user=user,
                local_site=local_site,
                status='P',
                has_local_sites_in_db=local_sites_in_db))

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'group': 'devgroup',
                    'view': 'to-group',
                })

        self.assertEqual(response.status_code, 200)

    def _test_to_group_with_unjoined_private_group(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for Incoming -> To Group with unjoined private groups.

        This accesses the Dashboard, checking the Incoming -> To Group view
        with unjoined private groups for queries and expected datagrid row
        results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        if local_sites_in_db:
            local_site_q = Q(local_site=local_site)
        else:
            local_site_q = Q()

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        group = self.create_review_group(
            name='new-private',
            invite_only=True,
            local_site=local_site)

        review_request = self.create_review_request(
            summary='Test 1',
            publish=True,
            local_site=local_site,
            local_id=1)
        review_request.target_groups.add(group)

        equeries = get_http_request_start_equeries(
            user=user,
            local_site=local_site)
        equeries += get_user_local_site_profile_equeries(
            user=user,
            profile=profile,
            local_site=local_site)
        equeries += [
            {
                'model': Group,
                'where': (Q(name='devgroup') &
                          local_site_q),
            },
        ]

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'group': 'devgroup',
                    'view': 'to-group',
                })

        self.assertEqual(response.status_code, 404)

    def _test_with_all_columns(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for checking all available columns.

        This accesses the Dashboard, enabling all columns and checking the
        resutling queries.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        profile.starred_groups.clear()

        review_requests = []
        diffset_histories = []

        review_request_pks: List[int] = []

        for i in range(10):
            review_request = self.create_review_request(
                summary='Test %s' % (i + 1),
                local_site=local_site,
                local_id=i + 1,
                publish=True)

            if i < 5:
                review_request.target_people.add(user)
                diffset_histories.append(review_request.diffset_history)
                review_requests.append(review_request)
                review_request_pks.append(review_request.pk)

        # This will be ordered with newest at the top.
        review_request_pks.reverse()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            extra=self._build_extra(user=user,
                                    include_new_review_count=True,
                                    include_mycomments=True,
                                    include_publicreviewcount=True),
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=review_request_pks,
            include_star_column=True,
            include_to_me_column=True,
            column_data_select_related={
                'diffset_history',
                'repository',
            },
            rows_q_result=get_review_requests_to_user_q(
                user=user,
                to_user=user,
                to_user_profile=profile,
                local_site=local_site,
                has_local_sites_in_db=local_sites_in_db,
                status='P'))

        equeries += [
            {
                '__note__': 'Fetch the diffsets for each review request',
                'model': DiffSet,
                'where': Q(history__in=diffset_histories),
            },
            {
                '__note__': 'Fetch the target groups for each review request',
                'join_types': {
                    'reviews_reviewrequest_target_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'tables': {
                    'reviews_group',
                    'reviews_reviewrequest_target_groups',
                },
                'extra': {
                    '_prefetch_related_val_reviewrequest_id': (
                        '"reviews_reviewrequest_target_groups".'
                        '"reviewrequest_id"',
                        []
                    ),
                },
                'where': Q(review_requests__in=review_requests),
            },
            {
                '__note__': 'Fetch the target users for each review request',
                'join_types': {
                    'reviews_reviewrequest_target_people': 'INNER JOIN',
                },
                'model': User,
                'num_joins': 1,
                'tables': {
                    'auth_user',
                    'reviews_reviewrequest_target_people',
                },
                'extra': {
                    '_prefetch_related_val_reviewrequest_id': (
                        '"reviews_reviewrequest_target_people".'
                        '"reviewrequest_id"',
                        []
                    ),
                },
                'where': Q(directed_review_requests__in=review_requests),
            },
        ]

        column_ids = sorted(
            _column.id
            for _column in DashboardDataGrid.get_columns()
        )

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'columns': ','.join(column_ids),
                    'view': 'incoming',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 5)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 5')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 4')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[3]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[4]['object'].summary, 'Test 1')

        self.assertEqual(
            [
                _stateful_column.id
                for _stateful_column in datagrid.columns
            ],
            column_ids)

    def _test_show_archived(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for showing archived review requests.

        This accesses the Dashboard with various archived review request
        states and filters.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        dashboard_url = self.get_datagrid_url(local_site=local_site)

        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        profile.starred_groups.clear()

        if local_site:
            local_site.users.add(user)

        visible = self.create_review_request(
            local_site=local_site,
            local_id=1,
            summary='Test 1',
            publish=True)
        archived = self.create_review_request(
            local_site=local_site,
            local_id=2,
            summary='Test 2',
            publish=True)
        muted = self.create_review_request(
            local_site=local_site,
            local_id=3,
            summary='Test 3',
            publish=True)

        visible.target_people.add(user)
        archived.target_people.add(user)
        muted.target_people.add(user)

        self.client.get(visible.get_absolute_url())
        self.client.get(archived.get_absolute_url())
        self.client.get(muted.get_absolute_url())

        visit = ReviewRequestVisit.objects.get(user__username=user,
                                               review_request=archived.pk)
        visit.visibility = ReviewRequestVisit.ARCHIVED
        visit.save(update_fields=('visibility',))

        visit = ReviewRequestVisit.objects.get(user__username=user,
                                               review_request=muted.pk)
        visit.visibility = ReviewRequestVisit.MUTED
        visit.save(update_fields=('visibility',))

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=[
                muted.pk,
                archived.pk,
                visible.pk,
            ],
            include_archived=True,
            rows_q_result=get_review_requests_to_user_q(
                user=user,
                to_user=user,
                to_user_profile=profile,
                local_site=local_site,
                has_local_sites_in_db=local_sites_in_db,
                status='P'))

        with assert_queries(equeries):
            response = self.client.get(
                dashboard_url,
                {
                    'show-archived': '1',
                    'view': 'incoming',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 3)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 1')

        response = self.client.get(
            dashboard_url,
            {
                'show-archived': '0',
                'view': 'incoming',
            })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 1)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 1')

        self.client.logout()
        self.client.login(username='grumpy', password='grumpy')
        user = User.objects.get(username='grumpy')

        if local_site:
            local_site.users.add(user)

        visible.target_people.add(user)
        archived.target_people.add(user)
        muted.target_people.add(user)

        self.client.get(visible.get_absolute_url())
        self.client.get(archived.get_absolute_url())
        self.client.get(muted.get_absolute_url())

        response = self.client.get(dashboard_url, {'show-archived': '1'})
        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 3)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 1')

        response = self.client.get(
            dashboard_url,
            {
                'show-archived': '0',
                'view': 'incoming',
            })
        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 3)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 1')

    def _test_archived_with_null_extra_data(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for archived review requests with null extra_data.

        This accesses the Dashboard with archived review requests containing
        null extra_data.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        # We encountered a bug where the archived state in the dashboard was
        # assuming that Profile.extra_data was always a dictionary. In modern
        # versions of Review Board, the default value for that field is an
        # empty dict, but old versions defaulted it to None. This test verifies
        # that the bug is fixed.
        archived = self.create_review_request(
            local_site=local_site,
            local_id=1,
            summary='Test 1',
            publish=True)

        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')

        profile = user.get_profile()
        profile.extra_data = None
        profile.save(update_fields=('extra_data',))
        profile.starred_groups.clear()

        archived.target_people.add(user)

        self.client.get(archived.get_absolute_url())

        visit = ReviewRequestVisit.objects.get(user__username=user,
                                               review_request=archived.id)
        visit.visibility = ReviewRequestVisit.ARCHIVED
        visit.save(update_fields=('visibility',))

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=[],
            rows_q_result=get_review_requests_to_user_q(
                user=user,
                to_user=user,
                to_user_profile=profile,
                local_site=local_site,
                has_local_sites_in_db=local_sites_in_db,
                status='P'))

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'show-archived': '0',
                    'view': 'incoming',
                })

        self.assertEqual(response.status_code, 200)

    def _test_sidebar(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for the sidebar.

        This accesses the Dashboard and checks the queries and results in
        the sidebar.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        if local_sites_in_db:
            local_site_q = Q(local_site=local_site)
        else:
            local_site_q = Q()

        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        user.review_groups.clear()

        # Create all the test data.
        devgroup = self.create_review_group(
            local_site=local_site,
            name='devgroup')
        devgroup.users.add(user)

        privgroup = self.create_review_group(
            local_site=local_site,
            name='privgroup')
        privgroup.users.add(user)

        self.create_review_request(
            local_site=local_site,
            local_id=1,
            submitter=user,
            publish=True)

        review_request2 = self.create_review_request(
            local_site=local_site,
            local_id=2,
            submitter='grumpy')
        draft = ReviewRequestDraft.create(review_request2)
        draft.target_people.add(user)
        review_request2.publish(review_request2.submitter)

        review_request3 = self.create_review_request(
            local_site=local_site,
            local_id=3,
            submitter='grumpy')
        draft = ReviewRequestDraft.create(review_request3)
        draft.target_groups.add(devgroup)
        review_request3.publish(review_request3.submitter)

        review_request4 = self.create_review_request(
            local_site=local_site,
            local_id=4,
            submitter='grumpy')
        draft = ReviewRequestDraft.create(review_request4)
        draft.target_groups.add(privgroup)
        review_request4.publish(review_request4.submitter)
        profile.star_review_request(review_request4)

        # Join a group.
        devgroup.users.add(user)

        # Star some groups.
        profile.star_review_group(devgroup)
        profile.star_review_group(privgroup)

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=[
                review_request4.pk,
                review_request3.pk,
                review_request2.pk,
            ],
            sidebar_equeries=[
                {
                    '__note__': (
                        "Fetch the list of a user's starred review groups"
                    ),
                    'join_types': {
                        'accounts_profile_starred_groups': 'INNER JOIN',
                    },
                    'model': Group,
                    'num_joins': 1,
                    'only_fields': {
                        'id',
                        'incoming_request_count',
                        'name',
                    },
                    'tables': {
                        'accounts_profile_starred_groups',
                        'reviews_group',
                    },
                    'where': (Q(starred_by__id=profile.pk) &
                              ~Q(pk__in={devgroup.pk, privgroup.pk}) &
                              Q(local_site_q)),
                },
            ],
            rows_q_result=get_review_requests_to_user_q(
                user=user,
                to_user=user,
                to_user_profile=profile,
                local_site=local_site,
                has_local_sites_in_db=local_sites_in_db,
                target_groups=[
                    devgroup,
                    privgroup,
                ],
                status='P'))

        # Now load the dashboard and get the sidebar items.
        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site),
                {
                    'view': 'incoming',
                })

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid

        sidebar_items = datagrid.sidebar_items
        self.assertEqual(len(sidebar_items), 3)

        # Test the "Overview" section.
        section = sidebar_items[0]
        self.assertEqual(str(section.label), 'Overview')

        # Test the "Outgoing" section.
        section = sidebar_items[1]
        self.assertEqual(str(section.label), 'Outgoing')
        self.assertEqual(len(section.items), 2)
        self.assertEqual(str(section.items[0].label), 'All')
        self.assertEqual(section.items[0].count, 1)
        self.assertEqual(str(section.items[1].label), 'Open')
        self.assertEqual(section.items[1].count, 1)

        # Test the "Incoming" section.
        section = sidebar_items[2]
        self.assertEqual(str(section.label), 'Incoming')
        self.assertEqual(len(section.items), 5)
        self.assertEqual(str(section.items[0].label), 'Open')
        self.assertEqual(section.items[0].count, 3)
        self.assertEqual(str(section.items[1].label), 'To Me')
        self.assertEqual(section.items[1].count, 1)
        self.assertEqual(str(section.items[2].label), 'Starred')
        self.assertEqual(section.items[2].count, 1)
        self.assertEqual(str(section.items[3].label), 'devgroup')
        self.assertEqual(section.items[3].count, 1)
        self.assertEqual(str(section.items[4].label), 'privgroup')
        self.assertEqual(section.items[4].count, 1)

    def _build_datagrid_equeries(
        self,
        *,
        user: User,
        profile: Profile,
        review_request_pks: List[int],
        rows_q_result: ExpectedQResult,
        extra: Dict[str, Any] = {},
        load_state_equeries: ExpectedQueries = [],
        sidebar_equeries: ExpectedQueries = [],
        local_site: Optional[LocalSite] = None,
        local_sites_in_db: bool = False,
        include_archived: bool = False,
        include_star_column: bool = False,
        include_to_me_column: bool = False,
        column_data_select_related: Set[str] = set(),
    ) -> ExpectedQueries:
        """Return expected queries for viewing the datagrid.

        This assumes that the user has access to the datagrid, and that any
        cacheable state is already cached.

        Version Added:
            5.0.7

        Args:
            user (django.contrib.auth.models.User):
                The user accessing the datagrid.

            profile (reviewboard.accounts.models.Profile):
                The user's profile.

            review_request_pks (list of int):
                The list of review request IDs that are expected to be
                listed, in result order.

            rows_q_result (dict):
                The result for a Q-expression for the datagrid rows queryset.

            extra (dict, optional):
                Extra queries to add to the rows queryset.

            load_state_equeries (list of dict, optional):
                Additional expected queries to add to the datagrid
                state-loading phase.

            sidebar_equeries (list of dict, optional):
                Additional expected queries to add to the sidebar-rendering
                phase.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site that's being accessed, if any.

            local_sites_in_db (bool, optional):
                Whether the database contains any Local Sites.

            include_archived (bool, optional):
                Whether archived review requests will be returned in results.

            include_star_column (bool, optional):
                Whether the star column will be included in the list of
                enabled dashboard columns.

            include_to_me_column (bool, optional):
                Whether the To Me column will be included in the list of
                enabled dashboard columns.

            column_data_select_related (set of str, optional):
                Any additional relations included when selecting related
                models for the columns.

        Returns:
            list of dict:
            The list of expected queries.
        """
        if local_sites_in_db:
            local_site_q = Q(local_site=local_site)
        else:
            local_site_q = Q()

        if not extra:
            extra = self._build_extra(user=user,
                                      include_new_review_count=True,
                                      include_mycomments=True)

        rows_q = rows_q_result['q']
        rows_q_tables = rows_q_result['tables']
        rows_q_num_joins = len(rows_q_tables) - 1
        rows_q_join_types = rows_q_result.get('join_types', {})
        rows_q_subqueries = rows_q_result.get('subqueries', [])
        next_subquery_i = len(rows_q_subqueries) + 1

        equeries = get_http_request_start_equeries(
            user=user,
            local_site=local_site)
        equeries += get_user_local_site_profile_equeries(
            user=user,
            profile=profile,
            local_site=local_site)
        equeries += load_state_equeries
        equeries += rows_q_result.get('prep_equeries', [])
        equeries += [
            {
                '__note__': "Fetch the list of a user's review groups",
                'join_types': {
                    'reviews_group_users': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'only_fields': {
                    'id',
                    'incoming_request_count',
                    'name',
                },
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': (Q(users__id=user.pk) &
                          Q(local_site_q)),
            },
        ]
        equeries += sidebar_equeries
        equeries += [
            {
                '__note__': 'Update datagrid state on the user profile',
                'type': 'UPDATE',
                'model': Profile,
                'where': Q(pk=user.pk),
            },
        ]

        if include_archived:
            equeries += [
                {
                    '__note__': (
                        'Fetch the number of items across all datagrid pages '
                        'with archived review requests shown'
                    ),
                    'annotations': {'__count': Count('*')},
                    'join_types': rows_q_join_types,
                    'model': ReviewRequest,
                    'num_joins': rows_q_num_joins,
                    'subqueries': rows_q_subqueries,
                    'tables': rows_q_tables,
                    'where': (Q(rows_q) &
                              Q(local_site=local_site)),
                },
            ]

            if review_request_pks:
                equeries += [
                    {
                        '__note__': (
                            'Fetch the IDs of the items for one page with '
                            'archived review requests shown'
                        ),
                        'join_types': rows_q_join_types,
                        'model': ReviewRequest,
                        'num_joins': rows_q_num_joins,
                        'subqueries': rows_q_subqueries,
                        'tables': rows_q_tables,
                        'values_select': ('pk',),
                        'extra': extra,
                        'where': (Q(rows_q) &
                                  Q(local_site=local_site)),
                        'order_by': ('-last_updated',),
                        'distinct': True,
                        'limit': len(review_request_pks),
                    },
                ]
        else:
            equeries += [
                {
                    '__note__': (
                        'Fetch the number of items across all datagrid pages'
                    ),
                    'annotations': {'__count': Count('*')},
                    'join_types': rows_q_join_types,
                    'model': ReviewRequest,
                    'num_joins': rows_q_num_joins,
                    'tables': rows_q_tables,
                    'where': (Q(rows_q) &
                              ~Q(__Q__subquery__=next_subquery_i) &
                              Q(local_site=local_site)),

                    'subqueries': rows_q_subqueries + [
                        {
                            'model': ReviewRequestVisit,
                            'values_select': ('review_request_id',),
                            'where': (Q(user=user) &
                                      ~Q(visibility='V')),
                        },
                    ],
                },
            ]

            if review_request_pks:
                equeries += [
                    {
                        '__note__': 'Fetch the IDs of the items for one page',
                        'join_types': rows_q_join_types,
                        'model': ReviewRequest,
                        'num_joins': rows_q_num_joins,
                        'tables': rows_q_tables,
                        'values_select': ('pk',),
                        'extra': extra,
                        'where': (Q(rows_q) &
                                  ~Q(__Q__subquery__=next_subquery_i) &
                                  Q(local_site=local_site)),
                        'order_by': ('-last_updated',),
                        'distinct': True,
                        'limit': len(review_request_pks),

                        'subqueries': rows_q_subqueries + [
                            {
                                'model': ReviewRequestVisit,
                                'values_select': ('review_request_id',),
                                'where': (Q(user=user) &
                                          ~Q(visibility='V')),
                            },
                        ],
                    },
                ]

        if include_star_column:
            equeries += [
                {
                    '__note__': (
                        "Starred column: Fetch the IDs of the page's review "
                        "requests that are starred"
                    ),
                    'join_types': {
                        'accounts_profile_starred_review_requests':
                            'INNER JOIN',
                    },
                    'model': ReviewRequest,
                    'num_joins': 1,
                    'tables': {
                        'accounts_profile_starred_review_requests',
                        'reviews_reviewrequest',
                    },
                    'values_select': ('pk',),
                    'where': (Q(starred_by__id=profile.pk) &
                              Q(pk__in=review_request_pks)),
                },
            ]

        if include_to_me_column:
            equeries += [
                {
                    '__note__': (
                        "To Me column: Fetch the IDs of the page's review "
                        "requests that are targeting the user"
                    ),
                    'join_types': {
                        'reviews_reviewrequest_target_people': 'INNER JOIN',
                    },
                    'model': ReviewRequest,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequest',
                        'reviews_reviewrequest_target_people',
                    },
                    'values_select': ('pk',),
                    'where': (Q(target_people__id=user.pk) &
                              Q(pk__in=[5, 4, 3, 2, 1])),
                },
            ]

        if review_request_pks:
            if local_site is not None:
                equeries += [
                    {
                        '__note__': (
                            'Fetch the data for one page based on the IDs '
                            'on the Local Site'
                        ),
                        'model': ReviewRequest,
                        'select_related': column_data_select_related | {
                            'local_site',
                            'submitter',
                        },
                        'extra': extra,
                        'where': Q(pk__in=review_request_pks),
                    },
                ]
            else:
                equeries += [
                    {
                        '__note__': (
                            'Fetch the data for one page based on the IDs'
                        ),
                        'model': ReviewRequest,
                        'select_related': column_data_select_related | {
                            'submitter'
                        },
                        'extra': extra,
                        'where': Q(pk__in=review_request_pks),
                    },
                ]

        return equeries

    def _build_extra(
        self,
        *,
        user: User,
        include_new_review_count: bool = False,
        include_mycomments: bool = False,
        include_publicreviewcount: bool = False,
    ) -> Dict[str, Any]:
        """Return data for the extra queries for the datagrid querysets.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the request.

            include_new_review_count (bool, optional):
                Whether to include the ``new_review_count`` column information.

            include_mycomments (bool, optional):
                Whether to include the ``mycomments`` column information.

            include_publicreviewcount (bool, optional):
                Whether to include the ``publicreviewcount`` column
                information.

        Returns:
            dict:
            The resulting extra query dictionary.
        """
        user_id = user.pk

        extra: Dict[str, Any] = {
            'draft_summary': ("""
                SELECT reviews_reviewrequestdraft.summary
                  FROM reviews_reviewrequestdraft
                  WHERE reviews_reviewrequestdraft.review_request_id =
                        reviews_reviewrequest.id
            """, []),
            'visibility': (f"""
                SELECT accounts_reviewrequestvisit.visibility
                  FROM accounts_reviewrequestvisit
                 WHERE accounts_reviewrequestvisit.review_request_id =
                       reviews_reviewrequest.id
                   AND accounts_reviewrequestvisit.user_id = {user_id}
            """, []),
        }

        if include_new_review_count:
            extra['new_review_count'] = (f"""
                SELECT COUNT(*)
                  FROM reviews_review, accounts_reviewrequestvisit
                  WHERE reviews_review.public
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
                    AND accounts_reviewrequestvisit.review_request_id =
                        reviews_reviewrequest.id
                    AND accounts_reviewrequestvisit.user_id = {user_id}
                    AND reviews_review.timestamp >
                        accounts_reviewrequestvisit.timestamp
                    AND reviews_review.user_id != {user_id}
            """, [])

        if include_mycomments:
            extra.update({
                'mycomments_my_reviews': (f"""
                    SELECT COUNT(1)
                      FROM reviews_review
                      WHERE reviews_review.user_id = {user_id}
                        AND reviews_review.review_request_id =
                            reviews_reviewrequest.id
                """, []),
                'mycomments_private_reviews': (f"""
                    SELECT COUNT(1)
                      FROM reviews_review
                      WHERE reviews_review.user_id = {user_id}
                        AND reviews_review.review_request_id =
                            reviews_reviewrequest.id
                        AND NOT reviews_review.public
                """, []),
                'mycomments_shipit_reviews': (f"""
                    SELECT COUNT(1)
                      FROM reviews_review
                      WHERE reviews_review.user_id = {user_id}
                        AND reviews_review.review_request_id =
                            reviews_reviewrequest.id
                        AND reviews_review.ship_it
                """, []),
            })

        if include_publicreviewcount:
            extra['publicreviewcount_count'] = ("""
                SELECT COUNT(*)
                  FROM reviews_review
                 WHERE reviews_review.public
                   AND reviews_review.base_reply_to_id is NULL
                   AND reviews_review.review_request_id =
                       reviews_reviewrequest.id
            """, [])

        return extra



"""Unit tests for RootReviewResource."""

from datetime import timedelta
from django.utils import timezone
from djblets.testing.decorators import add_fixtures
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import review_list_mimetype
from reviewboard.webapi.tests.urls import get_root_review_list_url


class ResourceListTests(BaseWebAPITestCase):
    """Testing the RootReviewResource list APIs."""

    fixtures = ['test_users']
    sample_api_url = 'reviews/'
    resource = resources.root_review

    def compare_item(self, item_rsp, review):
        """Compare a review item to a review returned from the API.

        Args:
            item_rsp (dict):
                The serialized review returned from the API.

            review (reviewboard.reviews.models.Review):
                The review instance to compare to.
        Raises:
            AssertionError:
                The API response was not equivalent to the object.
        """
        self.assertEqual(item_rsp['id'], review.pk)
        self.assertEqual(item_rsp['ship_it'], review.ship_it)
        self.assertEqual(item_rsp['body_top'], review.body_top)
        self.assertEqual(item_rsp['body_bottom'], review.body_bottom)

        if review.body_top_rich_text:
            self.assertEqual(item_rsp['body_top_text_type'], 'markdown')
        else:
            self.assertEqual(item_rsp['body_top_text_type'], 'plain')

        if review.body_bottom_rich_text:
            self.assertEqual(item_rsp['body_bottom_text_type'], 'markdown')
        else:
            self.assertEqual(item_rsp['body_bottom_text_type'], 'plain')

        self.assertEqual(item_rsp['absolute_url'],
                         self.base_url + review.get_absolute_url())

    @webapi_test_template
    def test_root_review_is_same_as_canonical_review(self):
        """Testing the GET <URL>/ API returns a proper data type API"""
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, publish=True)

        rsp = self.api_get(get_root_review_list_url(), {},
                           expected_mimetype=review_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.compare_item(rsp[self.resource.list_result_key][0], review)

    @webapi_test_template
    def test_get_with_counts_only(self):
        """Testing the GET <URL>/?counts-only=1 API returns expected
        counts
        """
        review_request = self.create_review_request(publish=True)
        self.create_review(review_request, publish=True)
        self.create_review(review_request, publish=True)

        rsp = self.api_get(get_root_review_list_url(), {
            'counts-only': 1,
        }, expected_mimetype=review_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['count'], 2)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_get_without_public(self):
        """Testing the GET <URL>/ API returns both public and unpublished
        reviews that the requester has access to by default
        """
        reviews = self._setup_acl_tests()
        review1 = reviews[0]
        review3 = reviews[2]
        review4 = reviews[3]
        review7 = reviews[6]

        rsp = self.api_get(get_root_review_list_url(), {},
                           expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 4)
        self.compare_item(rsp_items[0], review1)
        self.compare_item(rsp_items[1], review3)
        self.compare_item(rsp_items[2], review4)
        self.compare_item(rsp_items[3], review7)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_get_with_public_true(self):
        """Testing the GET <URL>/?public=1 API returns only public reviews
        that the requester has access to
        """
        reviews = self._setup_acl_tests()
        review1 = reviews[0]
        review3 = reviews[2]
        review7 = reviews[6]

        rsp = self.api_get(get_root_review_list_url(), {
            'public': 1,
        }, expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 3)
        self.compare_item(rsp_items[0], review1)
        self.compare_item(rsp_items[1], review3)
        self.compare_item(rsp_items[2], review7)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_get_with_public_false(self):
        """Testing the GET <URL>/?public=0 API returns only unpublished reviews
        that the requester has access to
        """
        reviews = self._setup_acl_tests()
        review4 = reviews[3]

        rsp = self.api_get(get_root_review_list_url(), {
            'public': 0,
        }, expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], review4)

    @webapi_test_template
    def test_get_by_user_without_param(self):
        """Testing the GET <URL>/ API returns reviews that the requester has
        access to from all users when the user parameter is not included
        """
        reviews = self._setup_get_by_user_tests()
        review1 = reviews[0]
        review3 = reviews[2]

        rsp = self.api_get(get_root_review_list_url(), {},
                           expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 2)
        self.compare_item(rsp_items[0], review1)
        self.compare_item(rsp_items[1], review3)

    @webapi_test_template
    def test_get_by_user_with_param(self):
        """Testing the GET <URL>/?user=name API returns reviews by user"""
        reviews = self._setup_get_by_user_tests()
        review1 = reviews[0]

        # Testing that only reviews by the given user that the requester
        # has access to are returned.
        rsp = self.api_get(get_root_review_list_url(), {
            'user': 'user1',
        }, expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], review1)

        # Testing that no reviews are returned when the given user
        # doesn't exist.
        rsp = self.api_get(get_root_review_list_url(), {
            'user': 'user3',
        }, expected_mimetype=review_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 0)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_get_by_repo(self):
        """Testing the GET <URL>/ API returns only reviews from
        repositories that the requester has access to
        """
        reviews = self._setup_get_by_repo_tests()
        review1 = reviews[0]
        review2 = reviews[1]
        review3 = reviews[2]
        review5 = reviews[4]
        review6 = reviews[5]

        rsp = self.api_get(get_root_review_list_url(), {},
                           expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 5)
        self.compare_item(rsp_items[0], review1)
        self.compare_item(rsp_items[1], review2)
        self.compare_item(rsp_items[2], review3)
        self.compare_item(rsp_items[3], review5)
        self.compare_item(rsp_items[4], review6)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_get_by_repo_with_param(self):
        """Testing the GET <URL>/?repository=name API returns reviews by
        repository and only from repositories that the requester has access to
        """
        reviews = self._setup_get_by_repo_tests()
        review1 = reviews[0]

        # Testing that reviews from the given repository that the requester
        # has access to are returned.
        rsp = self.api_get(get_root_review_list_url(), {
            'repository': 'repo1',
        }, expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], review1)

        # Testing that no reviews are returned from repositories that the
        # requester does not have access to.
        rsp = self.api_get(get_root_review_list_url(), {
            'repository': 'repo4',
        }, expected_mimetype=review_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 0)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_get_by_review_group(self):
        """Testing the GET <URL>/ API returns reviews associated with
        any review groups that the requester has access to
        """
        # Review that the requester has access to from being in a public
        # review group that is targeted by the review request.
        group1 = self.create_review_group(name='group1', invite_only=False)
        group1.users.add(self.user)
        review_request1 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request1.target_groups.add(group1)
        review1 = self.create_review(review_request1, publish=True)

        # Review that the requester has access to from being in an invite-only
        # review group that is targeted by the review request.
        group2 = self.create_review_group(name='group2', invite_only=True)
        group2.users.add(self.user)
        review_request2 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request2.target_groups.add(group2)
        review2 = self.create_review(review_request2, publish=True)

        # Review that the requester has access to since there is a public
        # review group that is targeted by the review request.
        group3 = self.create_review_group(name='group3', invite_only=False)
        review_request3 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request3.target_groups.add(group3)
        review3 = self.create_review(review_request3, publish=True)

        # Review that the requester does not have access to since there is an
        # invite-only review group that is targeted by the review request.
        group4 = self.create_review_group(name='group4', invite_only=True)
        review_request4 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request4.target_groups.add(group4)
        self.create_review(review_request4, publish=True)

        rsp = self.api_get(get_root_review_list_url(), {},
                           expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 3)
        self.compare_item(rsp_items[0], review1)
        self.compare_item(rsp_items[1], review2)
        self.compare_item(rsp_items[2], review3)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_get_by_review_group_with_param(self):
        """Testing the GET <URL>/?review-group=name API returns reviews from
        users in the given group and only reviews that the requester has
        access to
        """
        # Review group that the requester is in.
        group1 = self.create_review_group(name='group1')
        group1.users.add(self.user)

        # User who is in the same review group.
        user1 = self.create_user('user1')
        group1.users.add(user1)

        # Public review by a user who's in the review group.
        review_request1 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review1 = self.create_review(review_request1,
                                     user=user1,
                                     publish=True)

        # Private review by a user who's in the review group.
        self.create_review(review_request1, user=user1, publish=False)

        # Public review by the requester.
        review3 = self.create_review(review_request1,
                                     user=self.user,
                                     publish=False)

        # Another review group.
        group2 = self.create_review_group(name='group2')

        # User who is in the other review group.
        user2 = self.create_user('user2')
        group2.users.add(user2)

        # Public review by a user who's in the other review group.
        review_request2 = self.create_review_request(publish=True,
                                                     create_repository=True)
        self.create_review(review_request2, user=user2, publish=True)

        rsp = self.api_get(get_root_review_list_url(), {
            'review-group': group1.name
        }, expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 2)
        self.compare_item(rsp_items[0], review1)
        self.compare_item(rsp_items[1], review3)

    @add_fixtures(['test_site', 'test_scmtools'])
    @webapi_test_template
    def test_get_by_local_site(self):
        """Testing the GET <URL>/ API returns only reviews from
        local sites that the requester has access to
        """
        # Review from a private local site that the user has access to.
        local_site1 = self.get_local_site(self.local_site_name)
        local_site1.users.add(self.user)
        repo1 = self.create_repository(local_site=local_site1)
        review_request1 = self.create_review_request(publish=True,
                                                     local_site=local_site1,
                                                     repository=repo1)
        review1 = self.create_review(review_request1, publish=True)

        # Review from a private local site that the user does not have
        # access to.
        local_site2 = self.get_local_site('local-site-2')
        repo2 = self.create_repository(local_site=local_site2)
        review_request2 = self.create_review_request(publish=True,
                                                     local_site=local_site2,
                                                     repository=repo2)
        self.create_review(review_request2, publish=True)

        # Review from a public local site.
        local_site3 = self.create_local_site('public-local-site', public=True)
        repo3 = self.create_repository(local_site=local_site3)
        review_request3 = self.create_review_request(publish=True,
                                                     local_site=local_site3,
                                                     repository=repo3)
        self.create_review(review_request3, publish=True)

        # Review from a global site.
        review_request4 = self.create_review_request(publish=True)
        review4 = self.create_review(review_request4, publish=True)

        # Testing that only reviews from the given local site are returned.
        rsp = self.api_get(
            get_root_review_list_url(local_site_name=local_site1.name),
            {},
            expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], review1)

        # Testing that the requester is not authorized to make requests
        # to a local site that they do not have access to.
        rsp = self.api_get(
            get_root_review_list_url(local_site_name=local_site2.name),
            {},
            expected_status=403)

        self.assertEqual(rsp['stat'], 'fail')

        # Testing that reviews from local sites do not leak into results
        # from the global site.
        rsp = self.api_get(get_root_review_list_url(), {},
                           expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], review4)

    @webapi_test_template
    def test_get_by_last_updated_from_includes_from_date(self):
        """Testing the GET <URL>/?last-updated-from=<date> API
        returns only reviews within the from date
        """
        now = timezone.now()
        review_request = self.create_review_request(publish=True)

        # Review created within the from date.
        review1 = self.create_review(review_request, publish=True,
                                     timestamp=(now + timedelta(hours=1)))

        # Review created on the from date.
        review2 = self.create_review(review_request, publish=True,
                                     timestamp=now)

        # Review created outside of the from date.
        self.create_review(review_request, publish=True,
                           timestamp=(now - timedelta(hours=1)))

        rsp = self.api_get(get_root_review_list_url(), {
            'last-updated-from': now.isoformat(),
        }, expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 2)
        self.compare_item(rsp_items[0], review2)
        self.compare_item(rsp_items[1], review1)

    @webapi_test_template
    def test_get_by_last_updated_to_includes_to_date(self):
        """Testing the GET <URL>/?last-updated-to=<date>
        API returns only reviews within the to date exclusively
        """
        now = timezone.now()
        review_request = self.create_review_request(publish=True)

        # Review created within the to date.
        review1 = self.create_review(review_request, publish=True,
                                     timestamp=(now - timedelta(hours=1)))

        # Review created on the to date.
        self.create_review(review_request, publish=True,
                           timestamp=now)

        # Review created outside of the to date.
        self.create_review(review_request, publish=True,
                           timestamp=(now + timedelta(hours=1)))

        rsp = self.api_get(get_root_review_list_url(), {
            'last-updated-to': now.isoformat(),
        }, expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], review1)

    @webapi_test_template
    def test_get_by_last_updated_to_and_from_simple_date_format(self):
        """Testing the GET
        <URL>/?last-updated-from=<date>&last-updated-to=<date>
        API returns only reviews within the from and to date
        """
        timestamp_from = timezone.now()
        timestamp_to = timezone.now() + timedelta(hours=1)
        review_request = self.create_review_request(publish=True)

        # Review created within the from and to dates.
        review = self.create_review(
            review_request,
            publish=True,
            timestamp=(timestamp_from + timedelta(minutes=30)))

        # Review created before the from date.
        self.create_review(review_request, publish=True,
                           timestamp=(timestamp_from - timedelta(minutes=30)))

        # Review created after the to date.
        self.create_review(review_request, publish=True,
                           timestamp=(timestamp_to + timedelta(minutes=30)))

        rsp = self.api_get(get_root_review_list_url(), {
            'last-updated-from': timestamp_from.isoformat(),
            'last-updated-to': timestamp_to.isoformat(),
        }, expected_mimetype=review_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], review)

    def _setup_acl_tests(self):
        """Sets up common review objects for access control tests.

        Returns:
            tuple of reviewboard.reviews.models.Review:
            A tuple of 7 reviews that were set up.
        """
        # Publicly-accessible published review request.
        review_request = self.create_review_request(publish=True)

        # Published review on a publicly-accessible review request.
        review1 = self.create_review(review_request, publish=True)

        # Unpublished review on a publicly-accessible review request.
        review2 = self.create_review(review_request, publish=False)

        # Published review owned by the requester on a publicly-accessible
        # review request.
        review3 = self.create_review(review_request,
                                     user=self.user,
                                     publish=True)

        # Unpublished review owned by the requester on a publicly-accessible
        # review request.
        review4 = self.create_review(review_request,
                                     user=self.user,
                                     publish=False)

        # Published review request from a private repository the requester
        # does not have access to.
        repo = self.create_repository(public=False)
        review_request_inaccessible = self.create_review_request(
            repository=repo,
            publish=True)

        # Published review on a private repository the requester does
        # not have access to.
        review5 = self.create_review(review_request_inaccessible, publish=True)

        # Unpublished review on a private repository the requester does
        # not have access to.
        review6 = self.create_review(review_request_inaccessible,
                                     publish=False)

        # An invite-only review group used to limit access for the following
        # review requests.
        group = self.create_review_group(invite_only=True)

        # Published review from a review request that has an invite-only review
        # group not accessible to the requester, but the requester has access
        # to through being a targeted reviewer.
        review_request_targeted = self.create_review_request(publish=True)
        review_request_targeted.target_groups.add(group)
        review_request_targeted.target_people.add(self.user)
        review7 = self.create_review(review_request_targeted, publish=True)

        # Published review from a review request that has an invite-only review
        # group not accessible to the requester, and that the requester does
        # not have access to because the requester is not listed as a target
        # reviewer.
        review_request_untargeted = self.create_review_request(publish=True)
        review_request_untargeted.target_groups.add(group)
        review8 = self.create_review(review_request_untargeted, publish=True)

        return (
            review1, review2, review3, review4,
            review5, review6, review7, review8
        )

    def _setup_get_by_user_tests(self):
        """Sets up common review objects for getting by user tests.

        Returns:
            tuple of reviewboard.reviews.models.Review:
            A tuple of 4 reviews that were set up.
        """
        review_request = self.create_review_request(publish=True)

        user1 = self.create_user(username='user1')
        review1 = self.create_review(review_request, publish=True, user=user1)
        review2 = self.create_review(review_request, publish=False, user=user1)

        user2 = self.create_user(username='user2')
        review3 = self.create_review(review_request, publish=True, user=user2)
        review4 = self.create_review(review_request, publish=False, user=user2)

        return review1, review2, review3, review4

    def _setup_get_by_repo_tests(self):
        """Sets up common review objects for getting by repositories tests.

        Returns:
            tuple of reviewboard.reviews.models.Review:
            A tuple of 7 reviews that were set up.
        """
        # Review from a public repository.
        repo1 = self.create_repository(name='repo1', public=True)
        review_request1 = self.create_review_request(publish=True,
                                                     repository=repo1)
        review1 = self.create_review(review_request1, publish=True)

        # Review from a private repository that the requester has
        # access to from being listed in the repository's users list.
        repo2 = self.create_repository(name='repo2', public=False)
        repo2.users.add(self.user)
        review_request2 = self.create_review_request(publish=True,
                                                     repository=repo2)
        review2 = self.create_review(review_request2, publish=True)

        # An invite-only review group that the requester has access to.
        group_accessible = self.create_review_group(invite_only=True)
        group_accessible.users.add(self.user)

        # Review from a private repository that the user has
        # access to through being a member of a targeted review group.
        repo3 = self.create_repository(name='repo3', public=False)
        repo3.review_groups.add(group_accessible)
        review_request3 = self.create_review_request(publish=True,
                                                     repository=repo3)
        review3 = self.create_review(review_request3, publish=True)

        # Review from a private repository that the requester does
        # not have access to.
        repo4 = self.create_repository(name='repo4', public=False)
        review_request4 = self.create_review_request(publish=True,
                                                     repository=repo4)
        review4 = self.create_review(review_request4, publish=True)

        # Review from a private repository that the requester has access
        # to through being a member of a targeted review group and
        # being listed on the repository's users list.
        repo5 = self.create_repository(name='repo5', public=False)
        repo5.review_groups.add(group_accessible)
        repo5.users.add(self.user)
        review_request5 = self.create_review_request(publish=True,
                                                     repository=repo5)
        review5 = self.create_review(review_request5, publish=True)

        # An invite-only review group that the requester does not have
        # access to.
        group_inaccessible = self.create_review_group(invite_only=True)

        # Review from a private repository that targets an invite-only review
        # group, but that the requester has access to from being listed in the
        # repository's users list.
        repo6 = self.create_repository(name='repo6', public=False)
        repo6.review_groups.add(group_inaccessible)
        repo6.users.add(self.user)
        review_request6 = self.create_review_request(publish=True,
                                                     repository=repo6)
        review6 = self.create_review(review_request6, publish=True)

        # Review from a private repository that targets an invite-only review
        # group and that the requester does not have access to.
        repo7 = self.create_repository(name='repo7', public=False)
        repo7.review_groups.add(group_inaccessible)
        review_request7 = self.create_review_request(publish=True,
                                                     repository=repo7)
        review7 = self.create_review(review_request7, publish=True)

        return (
            review1, review2, review3, review4,
            review5, review6, review7
        )



"""Management command to find large diffs in the database.

Version Added:
    5.0.3
"""

from __future__ import annotations

import argparse
import csv
import sys
from datetime import timedelta
from typing import Counter, Optional

from django.core.management.base import BaseCommand, CommandError
from django.utils import timezone
from django.utils.translation import gettext as _

from reviewboard.diffviewer.models import FileDiff
from reviewboard.reviews.models import ReviewRequest


class Command(BaseCommand):
    """Management command to find large diffs in the database.

    This can be useful when diagnosing performance problems on a Review Board
    server.

    Version Added:
        5.0.3
    """

    help = _(
        'Find large diffs stored in the database, for diagnosing performance '
        'problems.'
    )

    def add_arguments(
        self,
        parser: argparse.ArgumentParser,
    ) -> None:
        """Add arguments to the command.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser for the command.
        """
        parser.add_argument(
            '--min-size',
            type=int,
            default=100000,
            metavar='MIN_SIZE_BYTES',
            help=_(
                'Minimum diff or parent diff size to include in a result. '
                ' A review request is included if a diff meets --min-size '
                'or --min-files. Defaults to 100000 (100KB).'
            ))
        parser.add_argument(
            '--min-files',
            type=int,
            default=50,
            help=_(
                'Minimum number of files to include in a result. A review '
                'request is included if a diff meets --min-size or '
                '--min-files. Defaults to 50.'
            ))
        parser.add_argument(
            '--start-id',
            type=int,
            help=_(
                'Starting review request ID for the scan. Either --start-id '
                'or --num-days must be specified.'
            ))
        parser.add_argument(
            '--end-id',
            type=int,
            default=None,
            help=_(
                'Last review request ID for the scan. Defaults to the last '
                'ID in the database.'
            ))
        parser.add_argument(
            '--num-days',
            type=int,
            help=_(
                'Number of days back to scan for diffs. Either --num-days or '
                '--start-id must be specified.'
            ))
        parser.add_argument(
            '--noinput',
            '--no-input',
            action='store_false',
            dest='interactive',
            help=_(
                'Disable prompting for confirmation before performing the '
                'scan.'
            ))

    def handle(
        self,
        **options,
    ) -> None:
        """Handle the command.

        This will perform the large diff scan, based on the provided options.

        Args:
            **options (dict, unused):
                Options parsed on the command line.

        Raises:
            django.core.management.CommandError:
                There was an error performing a diff migration.
        """
        min_size: int = options['min_size']
        min_files: int = options['min_files']
        num_days: Optional[int] = options['num_days']
        start_id: Optional[int] = options['start_id']
        end_id: Optional[int] = options['end_id']

        if num_days is None and start_id is None:
            raise CommandError(_(
                'Either --num-days or --start-id must be specified.'
            ))

        queryset = (
            ReviewRequest.objects
            .only('diffset_history', 'last_updated', 'pk', 'submitter')
            .order_by()
        )

        if start_id is not None:
            if end_id is None:
                queryset = queryset.filter(pk__gte=start_id)
            else:
                queryset = queryset.filter(pk__range=(start_id, end_id))

        if num_days is not None:
            queryset = queryset.filter(
                last_updated__gte=timezone.now() - timedelta(days=num_days))

        # Check whether to confirm before performing the scan, in case there's
        # a larger number of review requests than expected.
        if options['interactive']:
            num_review_requests = queryset.count()

            answer = input(
                f'This will scan {num_review_requests} review requests. '
                f'Continue? [Y/n] '
            )

            if answer not in ('Y', 'y'):
                self.stderr.write('Canceling the scan.')
                sys.exit(1)

        csv_writer = csv.writer(self.stdout)
        csv_writer.writerow([
            'Review Request ID',
            'Last Updated',
            'User ID',
            'Max Files',
            'Max Diff Size',
            'Max Parent Diff Size',
            'Diffset ID for Max Files',
            'Diffset ID for Max Diff Size',
            'Diffset ID for Max Parent Diff Size',
        ])

        for review_request in queryset.all():
            # Each of these default to 0, so we can safely run max() later
            # in the event that a dictionary is otherwise empty.
            diff_size_by_diffset = Counter[int]({0: 0})
            parent_diff_size_by_diffset = Counter[int]({0: 0})
            num_files_by_diffset = Counter[int]({0: 0})

            filediffs = (
                FileDiff.objects
                .filter(diffset__history=review_request.diffset_history_id)
                .only('pk', 'diffset', 'diff_hash', 'parent_diff_hash')
                .prefetch_related('diff_hash', 'parent_diff_hash')
            )

            for filediff in filediffs:
                diffset_id = filediff.diffset_id
                num_files_by_diffset[diffset_id] += 1

                try:
                    diff_size_by_diffset[diffset_id] += \
                        len(filediff.diff_hash.content or '')
                except Exception:
                    pass

                try:
                    parent_diff_size_by_diffset[diffset_id] += \
                        len(filediff.parent_diff_hash.content or '')
                except Exception:
                    pass

            # Figure out the largest values and diffsets.
            max_diffset_id_by_diff, max_diff_size = \
                max(diff_size_by_diffset.items(),
                    key=lambda pair: pair[1])
            max_diffset_id_by_parent_diff, max_parent_diff_size = \
                max(parent_diff_size_by_diffset.items(),
                    key=lambda pair: pair[1])
            max_diffset_id_by_num_files, max_num_files = \
                max(num_files_by_diffset.items(),
                    key=lambda pair: pair[1])

            if (max_diff_size > min_size or
                max_parent_diff_size > min_size or
                max_num_files > min_files):
                csv_writer.writerow([
                    review_request.pk,
                    review_request.last_updated,
                    review_request.submitter_id,
                    max_num_files,
                    max_diff_size,
                    max_parent_diff_size,
                    max_diffset_id_by_num_files,
                    max_diffset_id_by_diff,
                    max_diffset_id_by_parent_diff,
                ])



"""Unit tests for the Perforce tool."""

from __future__ import annotations

import os
import shutil
import subprocess
import time
import unittest
from hashlib import md5
from tempfile import mkdtemp
from typing import TYPE_CHECKING, cast

try:
    import P4
    from P4 import P4Exception
except ImportError:
    P4 = None
    P4Exception = None

from django.conf import settings
from djblets.testing.decorators import add_fixtures
from djblets.util.filesystem import is_exe_in_path
from kgb import SpyAgency

from reviewboard.diffviewer.testing.mixins import DiffParserTestingMixin
from reviewboard.scmtools.core import PRE_CREATION
from reviewboard.scmtools.errors import (AuthenticationError,
                                         RepositoryNotFoundError,
                                         SCMError,
                                         UnverifiedCertificateError)
from reviewboard.scmtools.perforce import PerforceTool, STunnelProxy
from reviewboard.scmtools.tests.testcases import SCMTestCase
from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase

if TYPE_CHECKING:
    from typing import ClassVar

    from typing_extensions import Self


has_p4 = is_exe_in_path('p4')
has_p4d = is_exe_in_path('p4d')


if P4 is not None:
    class DummyP4(P4.P4):
        """A dummy wrapper around P4 that does not connect.

        This is used for certain tests that need to simulate connecting without
        actually talking to a server.
        """

        def connect(self) -> Self:
            """Connect to the server."""
            return self
else:
    DummyP4 = None  # type: ignore


@unittest.skipIf(P4 is None, 'The p4python module is not installed')
@unittest.skipIf(not has_p4, 'The p4 command line tool is not installed')
class BasePerforceTestCase(SpyAgency, SCMTestCase):
    """Base class for all Perforce tests.

    This will set up a local p4d server for use by tests which need to connect
    to a real server.
    """

    #: The local p4d process.
    p4d_process: ClassVar[subprocess.Popen[str]]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case class.

        Raises:
            RuntimeError:
                The local p4d failed to start.
        """
        super().setUpClass()

        p4_repo = os.path.join(os.path.dirname(__file__),
                               '..', 'testdata', 'p4_repo')

        local_repo_path = mkdtemp(prefix='rb-tests-perforce-')
        os.rmdir(local_repo_path)
        shutil.copytree(p4_repo, local_repo_path)
        cls.local_repo_path = local_repo_path

        if has_p4d:
            cls.p4d_process = subprocess.Popen(
                ['p4d', '-p', '61666', '-r', local_repo_path],
                text=True,
            )

            # Poll until `p4 info` succeeds.
            timeout = 5.0  # seconds
            poll_interval = 0.1  # seconds
            elapsed = 0.0

            while elapsed < timeout:
                try:
                    result = subprocess.run(
                        ['p4', '-p', 'localhost:61666', 'info'],
                        capture_output=True,
                        timeout=1.0,
                        check=True,
                    )
                    if result.returncode == 0:
                        break
                except (subprocess.TimeoutExpired,
                        subprocess.CalledProcessError):
                    pass

                time.sleep(poll_interval)
                elapsed += poll_interval
            else:
                # If we get here, we timed out.
                cls.p4d_process.terminate()
                raise RuntimeError('p4d failed to start within timeout period')

    @classmethod
    def tearDownClass(cls) -> None:
        """Tear down the test case class."""
        super().tearDownClass()

        if has_p4d:
            try:
                cls.p4d_process.terminate()
                cls.p4d_process.wait()
            finally:
                shutil.rmtree(cls.local_repo_path)


class PerforceTests(DiffParserTestingMixin, BasePerforceTestCase):
    """Unit tests for Perforce."""

    fixtures = ['test_scmtools']
    tool: PerforceTool

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        self.repository = self.create_repository(
            name='localhost',
            path='localhost:61666',
            username='guest',
            password='guest12345',
            encoding='none',
            tool_name='Perforce')
        self.tool = cast(PerforceTool, self.repository.get_scmtool())

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()

        shutil.rmtree(os.path.join(settings.SITE_DATA_DIR, 'p4'),
                      ignore_errors=True)

    def test_init_with_p4_client(self) -> None:
        """Testing PerforceTool.__init__ with p4_client"""
        self.repository.extra_data['p4_client'] = 'test-client'

        tool = PerforceTool(self.repository)
        self.assertIsInstance(tool.client.client_name, str)
        self.assertEqual(tool.client.client_name, 'test-client')

    def test_init_with_p4_client_none(self) -> None:
        """Testing PerforceTool.__init__ with p4_client=None"""
        self.repository.extra_data['p4_client'] = None

        tool = PerforceTool(self.repository)
        self.assertIsNone(tool.client.client_name)

    def test_init_without_p4_client(self) -> None:
        """Testing PerforceTool.__init__ without p4_client"""
        self.assertIsNone(self.tool.client.client_name)

    def test_init_with_p4_host(self) -> None:
        """Testing PerforceTool.__init__ with p4_host"""
        self.repository.extra_data['p4_host'] = 'test-host'

        tool = PerforceTool(self.repository)
        self.assertIsInstance(tool.client.p4host, str)
        self.assertEqual(tool.client.p4host, 'test-host')

    def test_init_with_p4_host_none(self) -> None:
        """Testing PerforceTool.__init__ with p4_host=None"""
        self.repository.extra_data['p4_host'] = None

        tool = PerforceTool(self.repository)
        self.assertIsNone(tool.client.p4host)

    def test_init_without_p4_host(self) -> None:
        """Testing PerforceTool.__init__ without p4_host"""
        self.assertIsNone(self.tool.client.p4host)

    def test_connect_sets_required_client_args(self) -> None:
        """Testing PerforceTool.connect sets required client args"""
        self.repository.username = 'test-user'
        self.repository.password = 'test-pass'
        self.repository.encoding = 'utf8'
        self.repository.extra_data['use_ticket_auth'] = False

        tool = PerforceTool(self.repository)
        p4 = DummyP4()
        client = tool.client
        client.p4 = p4

        # Note that P4 will use the native string type on each major version
        # of Python. We want to sanity-check that here.
        with client.connect():
            self.assertEqual(p4.exception_level, 1)

            self.assertIsInstance(p4.user, str)
            self.assertEqual(p4.user, 'test-user')

            self.assertIsInstance(p4.password, str)
            self.assertEqual(p4.password, 'test-pass')

            self.assertIsInstance(p4.charset, str)
            self.assertEqual(p4.charset, 'utf8')

            self.assertIsInstance(p4.port, str)
            self.assertEqual(p4.port, 'localhost:61666')

            # Perforce will set a default for the host and client. They'll
            # be the same. We don't care what they are, just that they're
            # equal and of the right string type, and not "none".
            self.assertIsInstance(p4.host, str)
            self.assertIsInstance(p4.client, str)
            self.assertEqual(p4.host.split('.')[0], p4.client)
            self.assertNotEqual(p4.client.lower(), 'none')

            # Perforce will set the ticket file to be in the user's home
            # directory. We don't care about the exact contents, and will
            # just look at the filename.
            self.assertIsInstance(p4.ticket_file, str)
            self.assertTrue(p4.ticket_file.endswith('.p4tickets'))

    def test_connect_sets_optional_client_args(self) -> None:
        """Testing PerforceTool.connect sets optional client args"""
        self.repository.extra_data.update({
            'use_ticket_auth': True,
            'p4_client': 'test-client',
            'p4_host': 'test-host',
        })

        tool = PerforceTool(self.repository)
        p4 = DummyP4()
        client = tool.client
        client.p4 = p4

        self.spy_on(client.check_refresh_ticket, call_original=False)

        # Note that P4 will use the native string type on each major version
        # of Python. We want to sanity-check that here.
        with client.connect():
            self.assertIsInstance(p4.client, str)
            self.assertEqual(p4.client, 'test-client')

            self.assertIsInstance(p4.host, str)
            self.assertEqual(p4.host, 'test-host')

            self.assertIsInstance(p4.ticket_file, str)
            self.assertTrue(p4.ticket_file.endswith(
                os.path.join('data', 'p4', 'p4tickets')))

    def test_run_worker_with_unverified_cert(self) -> None:
        """Testing PerforceTool.run_worker with unverified certificate"""
        self.repository.path = 'p4.example.com:1666'
        self.repository.username = 'test-user'
        self.repository.password = 'test-pass'
        self.repository.encoding = 'utf8'
        self.repository.extra_data['use_ticket_auth'] = False

        tool = PerforceTool(self.repository)
        p4 = DummyP4()
        client = tool.client
        client.p4 = p4

        fingerprint = \
            'A0:B1:C2:D3:E4:F5:6A:7B:8C:9D:E0:F1:2A:3B:4C:5D:6E:7F:A1:B2'

        err_msg = (
            f"The authenticity of '1.2.3.4' can't be established,\\n"
            f"this may be your first attempt to connect to this P4PORT.\\n"
            f"The fingerprint for the key sent to your client is\\n"
            f"{fingerprint}\\n"
            f"To allow connection use the 'p4 trust' command.\\n"
        )

        expected_msg = (
            f'The SSL certificate for this repository (hostname '
            f'"p4.example.com:1666", fingerprint "{fingerprint}") was not '
            f'verified and might not be safe. This certificate needs to be '
            f'verified before the repository can be accessed.'
        )

        with self.assertRaisesMessage(UnverifiedCertificateError,
                                      expected_msg):
            with client.run_worker():
                raise P4Exception(err_msg)

    def test_run_worker_with_unverified_cert_new(self) -> None:
        """Testing PerforceTool.run_worker with new unverified certificate"""
        self.repository.path = 'p4.example.com:1666'

        tool = PerforceTool(self.repository)
        p4 = DummyP4()
        client = tool.client
        client.p4 = p4

        fingerprint = \
            'A0:B1:C2:D3:E4:F5:6A:7B:8C:9D:E0:F1:2A:3B:4C:5D:6E:7F:A1:B2'

        err_msg = (
            f"The authenticity of '1.2.3.4:1666' can't be established,\\n"
            f"this may be your first attempt to connect to this P4PORT.\\n"
            f"The fingerprint for the key sent to your client is\\n"
            f"{fingerprint}\\n"
            f"To allow connection use the 'p4 trust' command.\\n"
        )

        expected_msg = (
            f'The SSL certificate for this repository (hostname '
            f'"p4.example.com:1666", fingerprint "{fingerprint}") was not '
            f'verified and might not be safe. This certificate needs to be '
            f'verified before the repository can be accessed.'
        )

        with self.assertRaisesMessage(UnverifiedCertificateError,
                                      expected_msg):
            with client.run_worker():
                raise P4Exception(err_msg)

    def test_run_worker_with_unverified_cert_changed_error(self) -> None:
        """Testing PerforceTool.run_worker with unverified certificate and
        cert changed error
        """
        self.repository.path = 'p4.example.com:1666'

        tool = PerforceTool(self.repository)
        p4 = DummyP4()
        client = tool.client
        client.p4 = p4

        fingerprint = \
            'A0:B1:C2:D3:E4:F5:6A:7B:8C:9D:E0:F1:2A:3B:4C:5D:6E:7F:A1:B2'

        err_msg = (
            f"******* WARNING P4PORT IDENTIFICATION HAS CHANGED! *******\\n"
            f"It is possible that someone is intercepting your connection\\n"
            f"to the Perforce P4PORT '1.2.3.4:1666'\\n"
            f"If this is not a scheduled key change, then you should "
            f"contact\\n"
            f"your Perforce administrator.\\n"
            f"The fingerprint for the mismatched key sent to your client is\\n"
            f"{fingerprint}\n"
            f"To allow connection use the 'p4 trust' command.\n"
        )

        expected_msg = (
            f'The SSL certificate for this repository (hostname '
            f'"p4.example.com:1666", fingerprint "{fingerprint}") was not '
            f'verified and might not be safe. This certificate needs to be '
            f'verified before the repository can be accessed.'
        )

        with self.assertRaisesMessage(UnverifiedCertificateError,
                                      expected_msg):
            with client.run_worker():
                raise P4Exception(err_msg)

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_changeset(self) -> None:
        """Testing PerforceTool.get_changeset"""
        desc = self.tool.get_changeset(4)

        assert desc is not None

        self.assertEqual(desc.changenum, 4)
        self.assertEqual(desc.files, ['//depot/model.py', '//depot/readme'])
        self.assertEqual(desc.summary, 'Make some changes.')

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_encoding(self) -> None:
        """Testing PerforceTool.get_changeset with a specified encoding"""
        repo = self.create_repository(
            path='localhost:61666',
            username='guest',
            tool_name='Perforce',
            encoding='utf8')
        tool = repo.get_scmtool()

        try:
            tool.get_changeset('4')

            self.fail('Expected an error about unicode-enabled servers.')
        except SCMError as e:
            # Our local p4d doesn't have unicode enabled. Getting this
            # error means we at least passed the charset through correctly
            # to the p4 client.
            self.assertTrue('clients require a unicode enabled server' in
                            str(e))

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_with_auth_error_bad_password(self) -> None:
        """Testing PerforceTool.get_changeset with an authentication error
        from a bad password
        """
        repo = self.create_repository(
            path='localhost:61666',
            tool_name='Perforce',
            username='samwise',
            password='bogus',
            encoding='none')
        tool = repo.get_scmtool()

        message = 'Password invalid.'

        with self.assertRaisesMessage(AuthenticationError, message):
            tool.get_changeset('4')

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_with_auth_error_missing_password(self) -> None:
        """Testing PerforceTool.get_changeset with an authentication error
        from a missing password
        """
        repo = self.create_repository(
            path='localhost:61666',
            tool_name='Perforce',
            username='samwise',
            encoding='none')
        tool = repo.get_scmtool()

        message = (
            'Perforce password (P4PASSWD) invalid or unset.'
        )

        with self.assertRaisesMessage(AuthenticationError, message):
            tool.get_changeset('4')

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_with_auth_error_expired_password(self) -> None:
        """Testing PerforceTool.get_changeset with an authentication error
        from an expired password
        """
        repo = self.create_repository(
            path='localhost:61666',
            tool_name='Perforce',
            username='pending-user',
            password='temp12345',
            encoding='none')
        tool = repo.get_scmtool()

        message = (
            'Your password has expired, please change your password.'
        )

        with self.assertRaisesMessage(AuthenticationError, message):
            tool.get_changeset('4')

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_changeset_bad_p4port(self) -> None:
        """Testing PerforceTool.get_changeset with a bad P4PORT"""
        repo = self.create_repository(
            name='localhost:1',
            path='localhost:1',
            tool_name='Perforce')

        tool = repo.get_scmtool()
        self.assertRaises(RepositoryNotFoundError,
                          lambda: tool.get_changeset('1'))

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_get_file(self) -> None:
        """Testing PerforceTool.get_file"""
        tool = self.tool

        content = tool.get_file('//depot/foo', PRE_CREATION)
        self.assertIsInstance(content, bytes)
        self.assertEqual(content, b'')

        content = tool.get_file('//depot/model.py', '4')
        self.assertIsInstance(content, bytes)
        self.assertEqual(md5(content).hexdigest(),
                         'd41d8cd98f00b204e9800998ecf8427e')

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_file_exists(self) -> None:
        """Testing PerforceTool.file_exists"""
        self.assertTrue(self.tool.file_exists(
            '//depot/model.py', '2'))

        self.assertFalse(self.tool.file_exists(
            '//depot/xxx-non-existent', '1'))

    def test_file_exists_with_pre_creation(self) -> None:
        """Testing PerforceTool.file_exists"""
        self.assertFalse(self.tool.file_exists('//depot/xxx-new-file',
                                               PRE_CREATION))

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_custom_host(self) -> None:
        """Testing Perforce client initialization with a custom P4HOST"""
        repo = self.create_repository(
            path='localhost:61666',
            username='guest',
            password='guest12345',
            tool_name='Perforce',
            encoding='none')
        repo.extra_data['p4_host'] = 'my-custom-host'

        tool = repo.get_scmtool()
        assert isinstance(tool, PerforceTool)

        with tool.client.connect():
            self.assertEqual(tool.client.p4.host, 'my-custom-host')

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_ticket_login(self) -> None:
        """Testing Perforce with ticket-based logins"""
        repo = self.create_repository(
            path='localhost:61666',
            tool_name='Perforce',
            username='samwise',
            password='bogus')
        repo.extra_data = {
            'use_ticket_auth': True,
        }

        tool = repo.get_scmtool()
        assert isinstance(tool, PerforceTool)

        client = tool.client
        self.assertTrue(client.use_ticket_auth)

        self.spy_on(client.get_ticket_status, call_fake=lambda *args: {
            'user': 'samwise',
            'expiration_secs': 100000,
        })

        self.spy_on(client.login, call_original=False)

        self.assertFalse(os.path.exists(os.path.join(
            settings.SITE_DATA_DIR, 'p4', 'p4tickets')))

        with client.connect():
            self.assertSpyNotCalled(client.login)
            self.assertEqual(client.p4.ticket_file,
                             os.path.join(settings.SITE_DATA_DIR, 'p4',
                                          'p4tickets'))

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_ticket_login_with_expiring_ticket(self) -> None:
        """Testing Perforce with ticket-based logins with ticket close to
        expiring
        """
        repo = self.create_repository(
            path='localhost:61666',
            tool_name='Perforce',
            username='samwise',
            password='bogus')
        repo.extra_data = {
            'use_ticket_auth': True,
        }

        tool = repo.get_scmtool()
        assert isinstance(tool, PerforceTool)

        client = tool.client
        self.assertTrue(client.use_ticket_auth)

        self.spy_on(client.get_ticket_status, call_fake=lambda *args: {
            'user': 'samwise',
            'expiration_secs': 99,
        })

        self.spy_on(client.login, call_original=False)

        with client.connect():
            self.assertIsNotNone(client.p4.ticket_file)
            self.assertSpyCalled(client.login)
            self.assertEqual(client.p4.ticket_file,
                             os.path.join(settings.SITE_DATA_DIR, 'p4',
                                          'p4tickets'))

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_ticket_login_with_no_valid_ticket(self) -> None:
        """Testing Perforce with ticket-based logins without a valid ticket
        """
        repo = self.create_repository(
            path='localhost:61666',
            tool_name='Perforce',
            username='samwise',
            password='bogus')
        repo.extra_data = {
            'use_ticket_auth': True,
        }

        tool = repo.get_scmtool()
        assert isinstance(tool, PerforceTool)

        client = tool.client
        self.assertTrue(client.use_ticket_auth)

        self.spy_on(client.get_ticket_status, call_fake=lambda *args: None)
        self.spy_on(client.login, call_original=False)

        with client.connect():
            self.assertSpyCalled(client.login)
            self.assertEqual(client.p4.ticket_file,
                             os.path.join(settings.SITE_DATA_DIR, 'p4',
                                          'p4tickets'))

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_ticket_login_with_different_user(self) -> None:
        """Testing Perforce with ticket-based logins with ticket for a
        different user
        """
        repo = self.create_repository(
            path='localhost:61666',
            tool_name='Perforce',
            username='samwise',
            password='bogus')
        repo.extra_data = {
            'use_ticket_auth': True,
        }

        tool = repo.get_scmtool()
        assert isinstance(tool, PerforceTool)

        client = tool.client
        self.assertTrue(client.use_ticket_auth)

        self.spy_on(client.get_ticket_status, call_fake=lambda *args: {
            'user': 'other-user',
            'expiration_secs': 100000,
        })

        self.spy_on(client.login, call_original=False)

        with client.connect():
            self.assertSpyCalled(client.login)
            self.assertEqual(client.p4.ticket_file,
                             os.path.join(settings.SITE_DATA_DIR, 'p4',
                                          'p4tickets'))

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    @add_fixtures(['test_site', 'test_users'])
    def test_ticket_login_with_local_site(self) -> None:
        """Testing Perforce with ticket-based logins with Local Sites"""
        repo = self.create_repository(
            path='localhost:61666',
            tool_name='Perforce',
            username='samwise',
            password='bogus',
            local_site=LocalSite.objects.get(name='local-site-1'))
        repo.extra_data = {
            'use_ticket_auth': True,
        }

        tool = repo.get_scmtool()
        assert isinstance(tool, PerforceTool)

        client = tool.client
        self.assertTrue(client.use_ticket_auth)

        self.spy_on(client.get_ticket_status, call_fake=lambda *args: {
            'user': 'samwise',
            'expiration_secs': 100000,
        })

        self.spy_on(client.login, call_original=False)

        with client.connect():
            self.assertSpyNotCalled(client.login)
            self.assertEqual(client.p4.ticket_file,
                             os.path.join(settings.SITE_DATA_DIR, 'p4',
                                          'local-site-1', 'p4tickets'))

    def test_empty_diff(self) -> None:
        """Testing Perforce empty diff parsing"""
        diff = b'==== //depot/foo/proj/README#2 ==M== /src/proj/README ====\n'

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'//depot/foo/proj/README',
            orig_file_details=b'2',
            modified_filename=b'/src/proj/README',
            modified_file_details=b'',
            data=diff)

    def test_binary_diff(self) -> None:
        """Testing Perforce binary diff parsing"""
        diff = (
            b'==== //depot/foo/proj/test.png#1 ==A== /src/proj/test.png '
            b'====\nBinary files /tmp/foo and /src/proj/test.png differ\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'//depot/foo/proj/test.png',
            orig_file_details=b'1',
            modified_filename=b'/src/proj/test.png',
            modified_file_details=b'',
            binary=True,
            data=diff)

    def test_deleted_diff(self) -> None:
        """Testing Perforce deleted diff parsing"""
        diff = (
            b'==== //depot/foo/proj/test.png#1 ==D== /src/proj/test.png '
            b'====\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'//depot/foo/proj/test.png',
            orig_file_details=b'1',
            modified_filename=b'/src/proj/test.png',
            modified_file_details=b'',
            deleted=True,
            data=diff)

    def test_moved_file_diff(self) -> None:
        """Testing Perforce moved file diff parsing"""
        diff = (
            b'Moved from: //depot/foo/proj/test.txt\n'
            b'Moved to: //depot/foo/proj/test2.txt\n'
            b'--- //depot/foo/proj/test.txt  //depot/foo/proj/test.txt#2\n'
            b'+++ //depot/foo/proj/test2.txt  01-02-03 04:05:06\n'
            b'@@ -1 +1,2 @@\n'
            b'-test content\n'
            b'+updated test content\n'
            b'+added info\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'//depot/foo/proj/test.txt',
            orig_file_details=b'2',
            modified_filename=b'//depot/foo/proj/test2.txt',
            modified_file_details=b'01-02-03 04:05:06',
            moved=True,
            insert_count=2,
            delete_count=1,
            data=diff)

    def test_moved_file_diff_no_changes(self) -> None:
        """Testing Perforce moved file diff parsing without changes"""
        diff = (
            b'==== //depot/foo/proj/test.png#5 ==MV== '
            b'//depot/foo/proj/test2.png ====\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'//depot/foo/proj/test.png',
            orig_file_details=b'5',
            modified_filename=b'//depot/foo/proj/test2.png',
            modified_file_details=b'',
            moved=True,
            data=diff)

    def test_empty_and_normal_diffs(self) -> None:
        """Testing Perforce empty and normal diff parsing"""
        diff1_text = (
            b'==== //depot/foo/proj/test.png#1 ==A== '
            b'/src/proj/test.png ====\n'
        )
        diff2_text = (
            b'--- test.c  //depot/foo/proj/test.c#2\n'
            b'+++ test.c  01-02-03 04:05:06\n'
            b'@@ -1 +1,2 @@\n'
            b'-test content\n'
            b'+updated test content\n'
            b'+added info\n'
        )
        diff = diff1_text + diff2_text

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'//depot/foo/proj/test.png',
            orig_file_details=b'1',
            modified_filename=b'/src/proj/test.png',
            modified_file_details=b'',
            data=diff1_text)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'//depot/foo/proj/test.c',
            orig_file_details=b'2',
            modified_filename=b'test.c',
            modified_file_details=b'01-02-03 04:05:06',
            insert_count=2,
            delete_count=1,
            data=diff2_text)

    def test_diff_file_normalization(self) -> None:
        """Testing perforce diff filename normalization"""
        parser = self.tool.get_parser(b'')
        self.assertEqual(parser.normalize_diff_filename('//depot/test'),
                         '//depot/test')

    def test_unicode_diff(self) -> None:
        """Testing Perforce diff parsing with unicode characters"""
        diff = (
            '--- tést.c  //depot/foo/proj/tést.c#2\n'
            '+++ tést.c  01-02-03 04:05:06\n'
            '@@ -1 +1,2 @@\n'
            '-tést content\n'
            '+updated test content\n'
            '+added info\n'
        ).encode('utf-8')

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename='//depot/foo/proj/tést.c'.encode('utf-8'),
            orig_file_details=b'2',
            modified_filename='tést.c'.encode('utf-8'),
            modified_file_details=b'01-02-03 04:05:06',
            insert_count=2,
            delete_count=1,
            data=diff)


class PerforceStunnelTests(BasePerforceTestCase):
    """Unit tests for Perforce running through stunnel.

    Out of the box, Perforce doesn't support any kind of encryption on its
    connections. The recommended setup in this case is to run an stunnel server
    on the perforce server which bounces SSL connections to the normal p4 port.
    One can then start an stunnel on their client machine and connect via a
    localhost: P4PORT.

    For these tests, we set up an stunnel server which will accept secure
    connections and proxy (insecurely) to the local perforce server. We can
    then tell the Perforce SCMTool to connect securely to stunnel.
    """

    fixtures = ['test_scmtools']

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        if not is_exe_in_path('stunnel'):
            raise unittest.SkipTest('stunnel is not installed')

        cert = os.path.join(os.path.dirname(__file__),
                            '..', 'testdata', 'stunnel.pem')
        self.proxy = STunnelProxy('localhost:61666')
        self.proxy.start_server(cert)

        # Find an available port to listen on
        path = f'stunnel:localhost:{self.proxy.port}'

        self.repository = self.create_repository(
            name='localhost - secure',
            path=path,
            username='guest',
            password='guest12345',
            encoding='none',
            tool_name='Perforce')

        self.tool = self.repository.get_scmtool()
        assert isinstance(self.tool, PerforceTool)

        self.tool.use_stunnel = True

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()

        self.proxy.shutdown()

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_changeset(self) -> None:
        """Testing PerforceTool.get_changeset with stunnel"""
        desc = self.tool.get_changeset('4')

        self.assertEqual(desc.changenum, 4)
        self.assertEqual(desc.files, ['//depot/model.py', '//depot/readme'])
        self.assertEqual(desc.summary, 'Make some changes.')

    @unittest.skipIf(not has_p4d,
                     'The p4d command line tool is not installed')
    def test_get_file(self) -> None:
        """Testing PerforceTool.get_file with stunnel"""
        file = self.tool.get_file('//depot/foo', PRE_CREATION)

        self.assertIsInstance(file, bytes)
        self.assertEqual(file, b'')

        file = self.tool.get_file('//depot/model.py', '4')

        self.assertEqual(md5(file).hexdigest(),
                         'd41d8cd98f00b204e9800998ecf8427e')


class PerforceAuthFormTests(TestCase):
    """Unit tests for PerforceTool's authentication form."""

    def test_fields(self) -> None:
        """Testing PerforceTool authentication form fields"""
        form = PerforceTool.create_auth_form()

        self.assertEqual(list(form.fields), ['username', 'password'])
        self.assertEqual(form['username'].help_text, '')
        self.assertEqual(form['username'].label, 'Username')
        self.assertEqual(form['password'].help_text, '')
        self.assertEqual(form['password'].label, 'Password')

    @add_fixtures(['test_scmtools'])
    def test_load(self) -> None:
        """Tetting PerforceTool authentication form load"""
        repository = self.create_repository(
            tool_name='Perforce',
            username='test-user',
            password='test-pass')

        form = PerforceTool.create_auth_form(repository=repository)
        form.load()

        self.assertEqual(form['username'].value(), 'test-user')
        self.assertEqual(form['password'].value(), 'test-pass')

    @add_fixtures(['test_scmtools'])
    def test_save(self) -> None:
        """Tetting PerforceTool authentication form save"""
        repository = self.create_repository(tool_name='Perforce')

        form = PerforceTool.create_auth_form(
            repository=repository,
            data={
                'username': 'test-user',
                'password': 'test-pass',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.username, 'test-user')
        self.assertEqual(repository.password, 'test-pass')


class PerforceRepositoryFormTests(TestCase):
    """Unit tests for PerforceTool's repository form."""

    def test_fields(self) -> None:
        """Testing PerforceTool repository form fields"""
        form = PerforceTool.create_repository_form()

        self.assertEqual(list(form.fields),
                         ['path', 'mirror_path', 'use_ticket_auth'])
        self.assertEqual(form['path'].help_text,
                         'The Perforce port identifier (P4PORT) for the '
                         'repository. If your server is set up to use SSL '
                         '(2012.1+), prefix the port with "ssl:". If your '
                         'server connection is secured with stunnel (2011.x '
                         'or older), prefix the port with "stunnel:".')
        self.assertEqual(form['path'].label, 'Path')
        self.assertEqual(form['mirror_path'].help_text,
                         'If provided, this path will be used instead for '
                         'all communication with Perforce.')
        self.assertEqual(form['mirror_path'].label, 'Mirror Path')
        self.assertEqual(form['use_ticket_auth'].help_text, '')
        self.assertEqual(form['use_ticket_auth'].label,
                         'Use ticket-based authentication')

    @add_fixtures(['test_scmtools'])
    def test_load(self) -> None:
        """Tetting PerforceTool repository form load"""
        repository = self.create_repository(
            tool_name='Perforce',
            path='example.com:123/cvsroot/test',
            mirror_path=':pserver:example.com:/cvsroot/test',
            extra_data={
                'use_ticket_auth': True,
            })

        form = PerforceTool.create_repository_form(repository=repository)
        form.load()

        self.assertEqual(form['path'].value(), 'example.com:123/cvsroot/test')
        self.assertEqual(form['mirror_path'].value(),
                         ':pserver:example.com:/cvsroot/test')
        self.assertTrue(form['use_ticket_auth'].value())

    @add_fixtures(['test_scmtools'])
    def test_save(self) -> None:
        """Tetting PerforceTool repository form save"""
        repository = self.create_repository(tool_name='Perforce')
        self.assertIsNone(repository.extra_data.get('use_ticket_auth'))

        form = PerforceTool.create_repository_form(
            repository=repository,
            data={
                'path': 'ssl:perforce.example.com:1666',
                'mirror_path': 'mirror.example.com:1666',
                'use_ticket_auth': True,
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.path, 'ssl:perforce.example.com:1666')
        self.assertEqual(repository.mirror_path, 'mirror.example.com:1666')
        self.assertTrue(repository.extra_data.get('use_ticket_auth'))



"""Unit tests for reviewboard.hostingsvcs.models.HostingServiceAccount.

Version Added:
    5.0.1
"""

from reviewboard.hostingsvcs.errors import MissingHostingServiceError
from reviewboard.hostingsvcs.github import GitHub
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.testing import TestCase


class HostingServiceAccountTests(TestCase):
    """Unit tests for HostingServiceAccount."""

    def test_service(self):
        """Testing HostingServiceAccount.service"""
        account = HostingServiceAccount.objects.create(
            service_name=GitHub.hosting_service_id,
            username='user1',
            visible=True)

        self.assertIsInstance(account.service, GitHub)

    def test_service_bad_service_name(self):
        """Testing HostingServiceAccount.service with a service name that does
        not match any service in the registry
        """

        account = HostingServiceAccount.objects.create(
            service_name='blah',
            username='user1',
            visible=True)
        message = ('The hosting service "blah" could not be loaded. An '
                   'administrator should ensure all necessary packages and '
                   'extensions are installed.')

        with self.assertRaisesMessage(MissingHostingServiceError, message):
            account.service



from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService


class FakeHostingService(BaseHostingService):
    """A hosting service that is not provided by Review Board.

    Fake hosting services are intended to be used to advertise for Beanbag,
    Inc.'s Power Pack extension.
    """

    hosting_service_id = None


class FakeAWSCodeCommitHostingService(FakeHostingService):
    name = 'AWS CodeCommit'
    supported_scmtools = ['Git']
    hosting_service_id = 'aws-codecommit'


class FakeBitbucketServerHostingService(FakeHostingService):
    name = 'Bitbucket Server'
    supported_scmtools = ['Git']
    hosting_service_id = 'bitbucket-server'


class FakeGitHubEnterpriseHostingService(FakeHostingService):
    name = 'GitHub Enterprise'
    supported_scmtools = ['Git']
    hosting_service_id = 'github-enterprise'


class FakeVisualStudioTeamServicesHostingService(FakeHostingService):
    name = 'VisualStudio.com'
    supported_scmtools = [
        'Team Foundation Server',
        'Team Foundation Server (git)',
    ]
    hosting_service_id = 'visual-studio-online'


FAKE_HOSTING_SERVICES = {
    'aws-codecommit': FakeAWSCodeCommitHostingService,
    'bitbucket-server': FakeBitbucketServerHostingService,
    'github-enterprise': FakeGitHubEnterpriseHostingService,
    'visual-studio-online': FakeVisualStudioTeamServicesHostingService,
}



"""Base backend support for Subversion clients."""

from __future__ import annotations

from abc import ABC, abstractmethod, abstractproperty
from contextlib import contextmanager
from typing import (Any, Callable, Dict, Iterator, Optional, Sequence,
                    TYPE_CHECKING, Tuple)
from urllib.parse import quote

from django.utils.translation import gettext as _
from typing_extensions import Final, NotRequired, TypeAlias, TypedDict

from reviewboard.scmtools.core import HEAD
from reviewboard.scmtools.errors import SCMError
from reviewboard.scmtools.svn import RawSSLTrustDict

if TYPE_CHECKING:
    from collections import OrderedDict
    from datetime import datetime
    from reviewboard.scmtools.core import RevisionID


class SVNDirEntry(TypedDict):
    """Information on an entry in a directory.

    Version Added:
        6.0
    """

    #: The path to the entry.
    #:
    #: Type:
    #:     str
    path: str

    #: The revision when this was created.
    #:
    #: Type:
    #:     str
    created_rev: str


class SVNLogEntry(TypedDict):
    """Information on a log entry from Subversion.

    The log entry can cover a file, directory, or property.

    Version Added:
        6.0
    """

    #: The revision of the entry.
    #:
    #: Type:
    #:     bytes or str
    revision: str

    #: The author of the entry.
    #:
    #: Type:
    #:     bytes or str
    author: str

    #: The date of the entry.
    #:
    #: Type:
    #:     datetime.datetime
    date: NotRequired[datetime]

    #: The commit message.
    #:
    #: Type:
    #:     bytes or str
    message: str


class SVNRepositoryInfoDict(TypedDict):
    """Information about a Subversion repository.

    This provides information available to consumers of the API.

    Version Added:
        6.0
    """

    #: The root URL of the configured repository.
    #:
    #: Type:
    #:     str
    root_url: str

    #: The full URL of the configured repository.
    #:
    #: Type:
    #:     str
    url: str

    #: The UUID of the repository.
    #:
    #: Type:
    #:     str
    uuid: str


class Client(ABC):
    """Base class for a Subversion client."""

    #: The default start revision for log entries.
    LOG_DEFAULT_START: Final[str] = 'HEAD'

    #: The default end revision for log entries.
    LOG_DEFAULT_END: Final[str] = '1'

    #: An optional Python module required for Subversion support.
    required_module: Optional[str] = None

    ######################
    # Instance variables #
    ######################

    #: The path to the Subversion configuration directory.
    #:
    #: Type:
    #:     str
    config_dir: str

    #: The path to the repository.
    #:
    #: Type:
    #:     str
    repopath: str

    def __init__(
        self,
        config_dir: str,
        repopath: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        """Initialize the client.

        Args:
            config_dir (str):
                The path to the Subversion configuration directory.

            repopath (str):
                The path to the repository.

            username (str, optional):
                The username for the repository.

            password (str, optional):
                The password for the repository.
        """
        self.config_dir = config_dir
        self.repopath = repopath

    @contextmanager
    def communicate(self) -> Iterator[None]:
        """Context manager for any code communicating with Subversion.

        This should be used any time code is ready to perform an operation
        on a Subversion repository. It will catch any errors from the client
        that need to be processed and allow the client to normalize them.

        Version Added:
            6.0

        Yields:
            The communication operation will run in this context.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error communicating with Subversion.
        """
        try:
            with self.communicate_hook():
                yield
        except SCMError:
            # This is already an explicit error. Raise this as normal.
            raise
        except Exception as e:
            # Allow the client to convert the error to a normalized form.
            raise self.normalize_error(e) from e

    @contextmanager
    def communicate_hook(self) -> Iterator[None]:
        """Hook for running a communication task.

        This is intended for unit tests only. All exceptions are
        pass-through.

        Version Added:
            6.0

        Yields:
            The communication operation will run in this context.
        """
        yield

    @abstractmethod
    def normalize_error(
        self,
        e: Exception,
    ) -> Exception:
        """Normalize an exception from the client.

        This will process the exception information and return an exception
        suitable for forwarding to the backend, as documented below.

        Args:
            e (Exception):
                The exception to normalize.

        Returns:
            Exception:
            The normalized exception.

        Raises:
            reviewboard.scmtools.errors.AuthenticationError:
                An authentication error communicating with the repository.

            reviewboard.scmtools.errors.SCMError:
                A general error communicating with the repository.
        """
        raise NotImplementedError

    @abstractmethod
    def set_ssl_server_trust_prompt(
        self,
        cb: SSLServerTrustPromptFunc,
    ) -> None:
        """Set a function for verifying a SSL certificate.

        Args:
            cb (callable):
                The function to call for verifying SSL certificates.
        """
        raise NotImplementedError

    @abstractmethod
    def get_file(
        self,
        path: str,
        revision: RevisionID = HEAD,
    ) -> bytes:
        """Return the contents of a file from the repository.

        This attempts to return the raw binary contents of a file from the
        repository, given a file path and revision.

        Args:
            path (str):
                The path to the file in the repository.

            revision (reviewboard.scmtools.core.Revision, optional):
                The revision to fetch. Subclasses should default this to
                :py:data:`HEAD`.

            *args (tuple, unused):
                Additional unused positional arguments.

            **kwargs (dict, unused):
                Additional unused keyword arguments.

        Returns:
            bytes:
            The returned file contents.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found in the repository.

            reviewboard.scmtools.errors.InvalidRevisionFormatError:
                The ``revision`` argument was in an invalid format.

            reviewboard.scmtools.errors.SCMError:
                An unexpected error was encountered with the repository.
        """
        raise NotImplementedError

    @abstractmethod
    def get_keywords(
        self,
        path: str,
        revision: RevisionID = HEAD,
    ) -> str:
        """Return a space-separated list of SVN keywords for a given path.

        Args:
            path (str):
                The path to the file or directory.

            revision (str):
                The revision containing the keywords.

        Returns:
            str:
            The resulting keywords.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error was encountered with the repository.

                This may be a more specific subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def get_log(
        self,
        path: str,
        *,
        start: Optional[str] = None,
        end: Optional[str] = None,
        limit: Optional[int] = None,
        discover_changed_paths: bool = False,
        limit_to_path: bool = False,
    ) -> Sequence[SVNLogEntry]:
        """Return log entries at the specified path.

        The log entries will appear ordered from most recent to least,
        with ``start`` being the most recent commit in the range.

        Args:
            path (str):
                The path to limit log entries to.

            start (str, optional):
                The start revision for the log.

                If not specified, clients must default this to "HEAD".

            end (str, optional):
                The end revision for the log.

                If not specified, clients must default this to "1".

            limit (int, optional):
                The maximum number of entries to return.

            discover_changed_paths (bool, optional):
                Whether to include information on changed paths in the
                results.

            limit_to_path (bool, optional):
                Limits results to ``path`` without factoring in history from
                any branch operations.

        Returns:
            list of SVNLogEntry:
            The list of resulting log entries.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error was encountered with the repository.

                This may be a more specific subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def list_dir(
        self,
        path: str,
    ) -> OrderedDict[str, SVNDirEntry]:
        """Return the directory contents of the specified path.

        The result will be an ordered dictionary of contents, mapping
        filenames or directory names with directory information.

        Args:
            path (str):
                The directory path.

        Returns:
            collections.OrderedDict:
            A dictionary mapping directory names as strings to
            :py:class:`SVNDirEntry` information.
        """
        raise NotImplementedError

    @abstractmethod
    def diff(
        self,
        revision1: str,
        revision2: str,
        path: Optional[str] = None,
    ) -> bytes:
        """Return a diff between two revisions.

        The diff will contain the differences between the two revisions,
        and may optionally be limited to a specific path.

        Args:
            revision1 (str):
                The first revision in the range.

            revision2 (str):
                The second revision in the range.

            path (str, optional):
                An optional path to limit the diff to.

        Returns:
            bytes:
            The resulting diff contents.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error was encountered with the repository.

                This may be a more specific subclass.
        """
        raise NotImplementedError

    @abstractproperty
    def repository_info(self) -> SVNRepositoryInfoDict:
        """Metadata about the repository.

        See :py:class:`SVNRepositoryInfoDict` for contents.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error was encountered with the repository.

                This may be a more specific subclass.
        """
        raise NotImplementedError

    def normalize_path(
        self,
        path: str,
    ) -> str:
        """Normalize a path to a file/directory for a request to Subversion.

        If the path is an absolute path beginning at the base of the
        repository, it will be returned as-is. Otherwise, it will be appended
        onto the repository path, with any leading ``/`` characters on the
        path removed.

        If appending the path, care will be taken to quote special characters
        like a space, ``#``, or ``?``, in order to ensure that they're not
        mangled. There are many characters Subversion does consider valid that
        would normally be quoted, so this isn't true URL quoting.

        All trailing ``/`` characters will also be removed.

        Args:
            path (str):
                The path to normalize.

        Returns:
            str:
            The normalized path.
        """
        if path.startswith(self.repopath):
            norm_path = path
        else:
            # Some important notes for the quoting below:
            #
            # 1) Subversion requires that we operate off of a URI-based
            #    repository path in order for file lookups to at all work, so
            #    we can be sure we're building a URI here. That means we're
            #    safe to quote.
            #
            # 2) This is largely being mentioned because the original
            #    contribution to fix a lookup issue here with special
            #    characters was written to be compatible with local file
            #    paths. Support for that is a pretty common assumption, but
            #    is unnecessary, so the code here is safe.
            #
            # 3) We can't rely on urllib's standard quoting behavior.
            #    completely. Subversion has a specific table of characters
            #    that must be quoted, and ones that can't be. There is enough
            #    we can leverage from urlquote's own table, but we need to
            #    mark several more as safe.
            #
            #    See the "svn_uri_char_validity" look up table and notes here:
            #
            #    https://github.com/apache/subversion/blob/trunk/subversion/libsvn_subr/path.c
            #
            # 4) file:// URLs don't allow non-printable characters (character
            #    codes < 32), while non-file:// URLs do. We don't want to
            #    trigger issues in Subversion (earlier versions assume this
            #    is our responsibility), so we validate here.
            #
            # 5) Modern Subversion seems to handle its own normalization now,
            #    from what we can tell. That might not always be true, though,
            #    and we need to support older versions, so we'll continue to
            #    maintain this going forward.
            if self.repopath.startswith('file:'):
                # Validate that this doesn't have any unprintable ASCII
                # characters or older versions of Subversion will throw a
                # fit.
                for c in path:
                    if 0 <= ord(c) < 32:
                        raise SCMError(
                            _('Invalid character code %(code)s found in '
                              'path %(path)r.')
                            % {
                                'code': ord(c),
                                'path': path,
                            })

            norm_path = '%s/%s' % (
                self.repopath,
                quote(path.lstrip('/'), safe="!$&'()*+,'-./:=@_~")
            )

        return norm_path.rstrip('/')

    @abstractmethod
    def accept_ssl_certificate(
        self,
        path: str,
        on_failure: Optional[AcceptCertificateFunc] = None,
    ) -> Dict[str, Any]:
        """Attempt to accept a SSL certificate.

        If the repository uses SSL, this method is used to determine whether
        the SSL certificate can be automatically accepted.

        If the cert cannot be accepted, the ``on_failure`` callback
        is executed.

        Args:
            path (str):
                The repository path.

            on_failure (callable, optional):
                A function to call if the certificate could not be accepted.

        Returns:
            dict:
            Serialized information on the certificate.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                There was an error accepting the certificate.
        """
        raise NotImplementedError


#: Type for a callback handle for accepting SSL certificates.
#:
#: Version Added:
#:     6.0
AcceptCertificateFunc: TypeAlias = Callable[
    [Client, Exception, str, Dict[str, Any]],
    None,
]


#: Type for a callback handle for prompting for SSL trust information.
#:
#: Version Added:
#:     6.0
SSLServerTrustPromptFunc: TypeAlias = Callable[
    [RawSSLTrustDict],
    Tuple[bool, int, bool],
]
